import copy
import logging

main_logging = logging.getLogger("ffxivcalc")
base_spell_logging = main_logging.getChild("Base_Spell")

from ffxivcalc.Jobs.ActionEnum import name_for_id
import math
from ffxivcalc.Jobs.PlayerEnum import JobEnum
from ffxivcalc.Jobs.PlayerEnum import RoleEnum
from ffxivcalc.helperCode.requirementHandler import failedRequirementEvent
from random import random, uniform
from ffxivcalc.helperCode.helper_math import roundDown
Lock = 0.75

class FailedToCast(Exception):#Exception called if a spell fails to cast
    pass


class buff:
    """
    This class is any buff given to a player. It contains the buff's value
    MultDPS : float -> Value of the multiplier
    name : str -> Name of the buff. 
    isDebuff : -> True if is a debuff.
    """
    def __init__(self, MultDPS : float, name : str = "Unnamed", isDebuff : bool = False):
        self.MultDPS = MultDPS #DPS multiplier of the buff
        self.name = name
        self.isDebuff = isDebuff

    def __str__(self) -> str:
        return self.name + " (" + str(int(round(self.MultDPS-1,2)*100)) + "%) " 
    
    def __repr__(self) -> str:
        return str(self)

    def isEqual(self, otherBuff):
        """This function checks equality between self and another buff.
        Equality is defined as same name and multDPS and isDebuff value.
        """
        return (self.MultDPS == otherBuff.MultDPS) and (self.name == otherBuff.name) and (self.isDebuff == otherBuff.isDebuff)

class buffHistory:
    """
    This class represents an interval of time in which a buff was present.
    """

    def __init__(self, StartTime : float, EndTime : float):
        self.StartTime = StartTime
        self.EndTime = EndTime

    def __str__(self) -> str:
        return "Buff start : " + str(self.StartTime) + " end : " + str(self.EndTime)

    def isUp(self, timeStamp : float) -> bool:
        """
        This function returns weither the buff was active under the given timeStamp
        timeStamp : float -> Timestamp in secconds
        """
        return timeStamp >= self.StartTime and timeStamp <= self.EndTime
    
class buffPercentHistory(buffHistory):
    """
    This class is a buffHistory that was a percent damage bonus. This class will also hold
    the percent bonus gained from the buff.
    """

    def __init__(self, StartTime : float, EndTime : float, PercentBonus : float):
        super().__init__(StartTime, EndTime)
        self.PercentBonus = PercentBonus    
                             # Used by Shadow to know if buff is trick attack
        self.isTrickAttack = False

    def __str__(self) -> str:
        return super().__str__() + " percentBonus : " + str(self.PercentBonus)

    def getPercentBonus(self) -> float:
        """
        This function returns the PercentBonus of this buff
        """
        return self.PercentBonus
    
class ZIPAction:
    """
    This class holds the information of an action's damage. It will be used to efficiently and quicly compute
    many runs with random crit and DH in order to get the DPS distribution of runs. In other words, ZIPActions
    are a "pre-baked" version of the normal Spell class.
    Damage (int) : Value corresponding to the pre computed damage value of the action. Doesn't include DH and Crit damage.
    CritChange (float) : Chance for the action to crit.
    CritMultiplier (float) : Crit multiplier of the action.
    AutoCritBonus (float) : Damage bonus from crit buff when auto crit.
    DHChance (float) : Chance for the action to DH.
    AutoDHBonus (float) : Damage bonus from direct hit when auto direct hit.
    """

    def __init__(self, Damage : int, CritChance : float, CritMultiplier : float, DHChance : float, auto_crit : bool = False, auto_dh : bool = False, AutoCritBonus : float = 1, AutoDHBonus : float = 1):
        self.Damage = Damage
        self.CritChance = CritChance
        self.CritMultiplier = CritMultiplier
        self.DHChance = DHChance
        self.auto_crit = auto_crit
        self.auto_dh = auto_dh
        self.AutoCritBonus = AutoCritBonus
        self.AutoDHBonus = AutoDHBonus

    def ComputeRandomDamage(self) -> int:
        """
        This function computes random damage of an action.
        """

        CritHit = (random() <= self.CritChance)
        DirectHit = ((random() <= self.DHChance))

        UniformDamage = math.floor(self.Damage * uniform(0.95, 1.05))
        CritDamage = math.floor(UniformDamage * (1 + self.CritMultiplier if CritHit else 1) * (self.AutoCritBonus if self.auto_crit else 1))
        DHDamage = math.floor(CritDamage * (1.25 if DirectHit else 1) * (self.AutoDHBonus if self.auto_dh else 1))
        return DHDamage

class PreBakedAction:
    """
    This class is similar to ZIPAction, but it has less preprocessing than a ZIPAction does.
    A PreBakedAction only has the given buffs in memory since the Stats of the player are not assumed
    constant when computing the damage.
    IsTank : bool -> If player is a tank. In which case f_MAIN_DAMAGE is different.
    MainStatPercentageBonus : float -> PercentageBonus of the MainStat (given by comp)
    HasPotionEffect : bool -> If the Action is under the effect of a tincture. Assumed to be Grade 8 HQ
    PercentageBonus : list(float) -> Bonus multiplier of the action
    CritBonus : float -> Crit bonus of the action
    DHBonus : float -> DH Bonus of the action
    type : int -> type of the damage
    timeStamp : float -> Timestamp of the action
    AutoCrit : bool -> If is an auto crit (true)
    AutoDH : bool -> if is an auto DH (true)
    isFromPet : bool -> True if a Pet PreBakedAction
    isGCD : bool -> True if the action is a GCD
    gcdLockTimer : float -> Time value for which the player cannot take another GCD action.
    spellDPSBuff : float -> Flat bonus applied on this action
    isConditionalAction : bool -> True if this action is a conditional Action.
    """

    def __init__(self, IsTank : bool, MainStatPercentageBonus : float, buffList : list,
                 TraitBonus : float, Potency : int, type : int, timeStamp : float, 
                 nonReducableStamp : float, reducableStamp : float, AutoCrit : bool = False, AutoDH : bool = False,
                 isFromPet : bool = False, isGCD : bool = False,gcdLockTimer : float = 0, spellDPSBuff : float = 1,
                 isConditionalAction : bool = False ):
        self.IsTank = IsTank
        self.MainStatPercentageBonus = MainStatPercentageBonus
        #self.HasPotionEffect = HasPotionEffect
        self.buffList = buffList # This holds all buff that are not raid buffs, since those can be affected by f_SPD. So RaidBuffs are in PercentageBonus
        self.TraitBonus = TraitBonus
        self.type = type
        self.timeStamp = timeStamp
        self.Potency = Potency
        self.gcdLockTimer = gcdLockTimer

        self.nonReducableStamp = nonReducableStamp
        self.reducableStamp = reducableStamp

        self.AutoCrit = AutoCrit
        self.AutoDH = AutoDH
        self.AutoCritBonus = 1
        self.AutoDHBonus = 1
        
        self.isFromPet = isFromPet
        self.isGCD = isGCD
        self.spellDPSBuff = spellDPSBuff

        self.CritBonus = 0
        self.DHBonus = 0
        self.PercentageBonus = []
        self.potionIsActive = False  # True if this action has a potion applied to it.

        self.isConditionalAction = isConditionalAction
        self.isWildfire = False # True if is wildfire. Have to consider different since does not crit or dh.

    def resetPercentageBonus(self):
        self.PercentageBonus = []

    def ComputeExpectedDamage(self, f_MAIN_DMG : float, f_WD : float, f_DET : float, f_TEN : float, f_SPD : float, f_CritRate : float, f_CritMult : float, f_DH : float, DHAuto : float):
        """
        This function is called to compute the damage of the action.
        This function requires all the values computed from the stat of the player
        These values can be computed using the Fight.ComputeFunctions logic.
        This function also returns Damage without crit and DH in order to facilitate the computation
        of random action damage in ComputeRandomDamage (which is computed afterward)
        n : int -> number of time for which the PreBakedAction will compute the random damage.
        """

        f_DET_DH = math.floor((f_DET + DHAuto) * 1000 ) / 1000


        # Hardcoded for now

        if self.isFromPet : f_WD = (132+math.floor(390*100/1000))/100

        Damage = 0
        if self.type == 0: # Type 0 is direct damage
            Damage = math.floor(math.floor(math.floor(math.floor(math.floor(self.Potency * f_MAIN_DMG) * (f_DET_DH if (self.AutoCrit and self.AutoDH) else f_DET)) * f_TEN ) *f_WD) * self.TraitBonus) # Player.Trait is trait DPS bonus
        elif self.type == 1: # Type 1 is magical DOT
            Damage = math.floor(math.floor(math.floor(math.floor(math.floor(math.floor(self.Potency * f_WD) * f_MAIN_DMG) * f_SPD) * f_DET) * f_TEN) * self.TraitBonus) + 1
        elif self.type == 2: # Type 2 is physical DOT
            Damage = math.floor(math.floor(math.floor(math.floor(math.floor(math.floor(self.Potency * f_MAIN_DMG) * f_DET) * f_TEN) * f_SPD) * f_WD) * self.TraitBonus) + 1
        elif self.type == 3: # Auto-attacks
            Damage = math.floor(math.floor(math.floor(math.floor(self.Potency * f_MAIN_DMG) * f_DET) * f_TEN) * f_SPD)
            Damage = math.floor(math.floor(Damage * math.floor(f_WD * (3/3) *100 )/100) * self.TraitBonus) # Player.Delay is assumed to be 3 for simplicity for now
        
        Damage = math.floor(Damage * self.spellDPSBuff)

        for buff in self.PercentageBonus:
            Damage = math.floor(Damage * buff)

                             # If is wildfire don't crit or dh so just return
        if self.isWildfire: return Damage, Damage

        
        auto_crit_bonus = (1 + self.CritBonus * f_CritMult) if self.AutoCrit else 1# Auto_crit bonus if buffed
        auto_dh_bonus = (1 + (self.DHBonus) * 0.25) if self.AutoDH else 1# Auto_DH bonus if buffed

        ExpectedDamage = math.floor(Damage *         (1 + ( (f_CritRate + self.CritBonus) if not self.AutoCrit else 1)  * f_CritMult))
        ExpectedDamage = math.floor(ExpectedDamage * (1 + ((f_DH + self.DHBonus) if not self.AutoDH else 1) * 0.25))
        ExpectedDamage = math.floor(ExpectedDamage * auto_crit_bonus)
        ExpectedDamage = math.floor(ExpectedDamage * auto_dh_bonus)

        base_spell_logging.debug("PreBakedAction has expected damage of " + str(ExpectedDamage) + " Potency :" + str(self.Potency) +" Trait : " + str(self.TraitBonus) + " critBonus : " + str(self.CritBonus) + " DHBonus : " + str(self.DHBonus))
        
        buffName = ""
        for buff in self.buffList:
            buffName += buff.name + " " + str(buff.MultDPS) + " : "
        base_spell_logging.debug("Active Buff : " + buffName + " PotionOn : " + str(self.potionIsActive))
        #base_spell_logging.debug(str((f_MAIN_DMG, f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH)))

        return ExpectedDamage, Damage
    
    def ComputeRandomDamage(self,Damage : int, f_CritRate : float, f_CritMult : float, f_DH : float) -> int:
        """
        This function computes random damage of a PreBakedAction. It uses the Damage value precomputed in the
        ComputeExpectedDamage in order to make the computation faster.
        Damage : int -> Damage value without Crit/DH.
        Relevant player values fr Crit/DH.
        """
                             # Checking if Critical and/or DH.
        CritHit = (random() <= (f_CritRate + self.CritBonus)) or self.AutoCrit
        DirectHit = ((random() <= (f_DH + self.DHBonus))) or self.AutoDH

        auto_crit_bonus = (1 + (self.CritBonus * f_CritMult)) if self.AutoCrit else 1# Auto_crit bonus if buffed
        auto_dh_bonus = (1 + (self.DHBonus * 0.25)) if self.AutoDH else 1# Auto_DH bonus if buffed

        UniformDamage = math.floor(Damage * uniform(0.95, 1.05))
        CritDamage = math.floor(UniformDamage * (1 + f_CritMult if CritHit else 1) * (self.AutoCritBonus if self.AutoCrit else 1))
        RandomDamage = math.floor(CritDamage * (1.25 if DirectHit else 1) * (self.AutoDHBonus if self.AutoDH else 1))
        RandomDamage = math.floor(RandomDamage * auto_crit_bonus)
        RandomDamage = math.floor(RandomDamage * auto_dh_bonus)

        return RandomDamage

class Spell:
    """
    This class is any Spell, it will have some subclasses to take Job similar spell, etc.
    """
    def __init__(self, id : int, GCD : bool, CastTime : float, RecastTime : float, Potency : int, ManaCost : int, Effect, Requirement, type : int = 0, aoe_fn = None, AOEHeal = False, TargetHeal = False):
        """
        Initialization of a Spell
        id : int -> id to identify the action
        GCD : bool -> True if the action is a GCD
        CastTime : float -> Cast time of the action
        RecastTime : float -> Recast time of the action
        Potency : int -> base potency of the action
        Manacost : int -> base manacost of the action
        Effect : function -> A function called upon the execution of the action which affects the player and the enemy.
        Requirement : (function -> Bool) -> function called upon the execution to verify if the action can be executed.
        type (int) : Type of the action. The types are Spell, Weaponskill and Ability. Type = 0 is ability, type = 1 is Spell, type = 2 is Weaponskill and type = 3 is a Limit Break
        aoe_fn (function) : Function that will be called (eeeeeeh)
        AOEHeal : bool -> True if the action is an AOE healing action
        TargetHeal : bool -> True if the action is a target healing action
        """

        if aoe_fn == None:
            def f():
                pass
            self.aoe_fn = f

        self.id = id
        self.GCD = GCD #True if GCD
        self.Potency = Potency
        self.ManaCost = ManaCost
        self.CastTime = CastTime
        self.RecastTime = RecastTime
        self.Effect = [Effect]
        self.Requirement = Requirement
        self.DPSBonus = 1
        self.TargetID = 0 #By default 0
        self.TargetPlayerObject = None # This is usually None, but will be set for actions that target.
        self.type = type 
        self.AOEHeal = AOEHeal
        self.TargetHeal = TargetHeal
        self.conditionalAction = False

    def __repr__(self):
        return f'Spell id : {self.id} - Class : {self.__class__}'

    def Cast(self, player, Enemy):
        """
        This function is called by the simulator when an action is ready to begin its casting. It checks for the requirement and apply all effect
        currently in the fight that can affect the action. It will lock the player into casting mode if necessary. The action is not executed here
        and is in a way being preapred to be executed. It will be checked if the action can be done.
        player : player -> player object casting
        Enemey : Enemy -> Enemy object on which the action is done.
        """

        if self.type == 3 : # If is a limit break
            return copy.deepcopy(self)

        tempSpell = copy.deepcopy(self)
        #Creating a tempSpell which will have its values changed according that what effect
        #the player and the enemy have
        #Will apply each effect the player currently has on the spell
        if self.id != -1: #id = -1 is WaitAbility, we don't want anything with that
            for Effect in player.EffectList:
                Effect(player, tempSpell)#Changes tempSpell
            for Effect in Enemy.EffectList:
                Effect(player, tempSpell)#Changes tempSpell
        #Checks if we meet the spell requirement

        # Round casting and recasting time :

        tempSpell.notRoundCastTime = tempSpell.CastTime
        tempSpell.notRoundRecastTime = tempSpell.RecastTime

        tempSpell.CastTime = roundDown(tempSpell.CastTime, 3)
        tempSpell.RecastTime = roundDown(tempSpell.RecastTime, 3)



        #Remove all effects that have to be removed

        for remove in player.EffectToRemove:
            player.EffectList.remove(remove) #Removing effect
        for add in player.EffectToAdd:
            player.EffectList.append(add)
        
        player.EffectToRemove = [] #Empty the remove list
        player.EffectToAdd = []

        for Requirement in tempSpell.Requirement:
            ableToCast, timeLeft = Requirement(player, tempSpell)

                             # I just realized that this logic has an issue. If a time based requirement fails, then the simulator checks
                             # if we can just wait. If it can it omits to check all other requirements. Effectively ignoring other requirements
                             # that could result in a simulation crash. Leaving this here so I don't forget to look deeper into it
                             # at a later point.

            if(not ableToCast): #Requirements return both whether it can be casted and will take away whatever value needs to be reduced to cast
                #Will check if timeLeft is within a margin, so we will just wait for it to come
                #timeLeft is the remaining time before the spell is available

                addInfo = "" if timeLeft <= 0 else "player had to wait for or would have to wait for " + str(timeLeft) + " seconds. GCDLock " + str(player.GCDLockTimer)

                fatal =  not(timeLeft <= player.CurrentFight.waitingThreshold and timeLeft > 0) and  (player.CurrentFight.RequirementOn)  # true if stops the simulation

                newFailedRequirementEvent = failedRequirementEvent(player.CurrentFight.TimeStamp, player.playerID, Requirement.__name__, addInfo, fatal) # Recording the event
                player.CurrentFight.failedRequirementList.append(newFailedRequirementEvent) # storing the event in memory

                log_str = ("FailedRequirementEvent, " + " , Timestamp : " + str(player.CurrentFight.TimeStamp)
                + " , PlayerID : " + str(player.playerID) + " , RequirementName : " + Requirement.__name__ + " , Fatal : " + str(fatal) + " , Info : " + addInfo)

                if fatal : base_spell_logging.critical(log_str) # if fatal makes the sim crash
                else : base_spell_logging.warning(log_str) # if not fatal doesn't crash the sim
                
                if not (player.CurrentFight.RequirementOn) : return tempSpell # If we do not care about requirement simply go on.
                elif timeLeft <= player.CurrentFight.waitingThreshold and timeLeft > 0: # If we care about requirement, we check if we can wait the allocated threshold. if we can we wait for it to come off cooldown.
                    # Limit of waiting for 1 sec
                    tempSpell = WaitAbility(timeLeft + 0.01)
                    player.ActionSet.insert(player.NextSpell, tempSpell)
                    return tempSpell #Makes the character wait
                    #Might remove some stuff tho, might have to check into that (for when effects are applied)
                
                player.CurrentFight.wipe = True # otherwise we stop the simulation 
                return tempSpell
        #Will make sure CastTime is at least Lock
        if tempSpell.id > 0 and tempSpell.id != 212 and tempSpell.CastTime < Lock : tempSpell.CastTime = 0 #id < 0 are special abilities like DOT, so we do not want them to be affected by that
        return tempSpell
        #Will put casting spell in player, and do damage/effect once the casting time is over


    def CastFinal(self, player, Enemy):

        """
        This function is called when an action is ready to be casted and apply its damage and effect.
        player : player -> player object casting
        Enemy : Enemy -> Enemy object on which the action is done.
        """
        if self.type == 3:
            pass # If limit break do not do effect
        else :
            for Effect in self.Effect:
                Effect(player, Enemy)#Put effects on Player and/or Enemy

        #This will include substracting the mana (it has been verified before that the mana was enough)
        minDamage, Damage, Heal = 0,0,0
        if self.AOEHeal or self.TargetHeal:
            type = 1
            if self.AOEHeal:
                             # Affects all players
                for gamer in player.CurrentFight.PlayerList:
                    Heal = 0 # player.CurrentFight.ComputeHealingFunction(player, self.Potency, gamer, 1, type, self)[0]
                    #base_spell_logging.debug(
                    #    "Timestamp : " + str(player.CurrentFight.TimeStamp) + " Player " + str(gamer.playerID) + "received a healing of min value : " + str(heal[0])
                    #)
                    gamer.HP += Heal
        else:
            type = 0 #Default value for type
            if isinstance(self, Auto_Attack):
                type = 3
            elif isinstance(self, DOTSpell): #Then dot
                #We have to figure out if its a physical dot or not
                if self.isPhysical: type = 2
                else: type = 1   


            #if player.CurrentFight.SavePreBakedAction:
            #                    # Adding to totalTimeNoFaster
            #    if self.GCD and self.RecastTime <= 1.5: # We check that the spellObj has recastTime lower than 1.5 and that it is not the last spell (since all those are insta cast)
            #        if  player.isLastGCD(player.NextSpell): # if last GCD, add CastTime
            #            player.totalTimeNoFaster += self.notRoundCastTime
            #        else:        # Else adding recastTime. 
            #            player.totalTimeNoFaster += self.notRoundRecastTime
            #    elif not self.GCD and player.isLastGCD(player.NextSpell) : 
            #                    # Is an oGCD
            #        player.totalTimeNoFaster += self.notRoundCastTime
            #    elif self.GCD and (player.RoleEnum == RoleEnum.Caster) and self.type == 2: # If is a weaponskill and has Spell speed. Only needed for RDM now
            #        if  player.isLastGCD(player.NextSpell): # if last GCD, add CastTime
            #            player.totalTimeNoFaster += self.CnotRoundastTime
            #        else:        # Else adding recastTime. 
            #            player.totalTimeNoFaster += self.notRoundRecastTime
            #    elif self.GCD and (player.RoleEnum == RoleEnum.Melee or player.RoleEnum == RoleEnum.Tank) and self.type == 1: # If is a spell and has skill speed
            #        if  player.isLastGCD(player.NextSpell): # if last GCD, add CastTime
            #            player.totalTimeNoFaster += self.notRoundCastTime
            #        else:        # Else adding recastTime. 
            #            player.totalTimeNoFaster += self.notRoundRecastTime

                                # If the action has 0 potency we skip the computation
                                # Note that this also means the action won't be added as a ZIPAction for the player.
                                # Also won't be added as a PreBakedAction. This can cause issue for waitAbility if scheduled at the end.
                                # The preBakedAction in that case will go until the last AA which might not be at the same time as waitAbility 
                                # finishes. This only affects end time and is not a super big issue. Probably just don't put waitAbility at the end?
            if self.type == 3:
                minDamage,Damage= player.CurrentFight.ComputeDamageFunction(player, self.Potency, Enemy, self.DPSBonus, type, self, 
                                                                            SavePreBakedAction = player.CurrentFight.SavePreBakedAction, 
                                                                            PlayerIDSavePreBakedAction = player.CurrentFight.PlayerIDSavePreBakedAction)    #Damage computation
            elif self.Potency != 0 : minDamage,Damage= player.CurrentFight.ComputeDamageFunction(player, self.Potency, Enemy, self.DPSBonus, type, self, SavePreBakedAction = player.CurrentFight.SavePreBakedAction, PlayerIDSavePreBakedAction = player.CurrentFight.PlayerIDSavePreBakedAction)    #Damage computation
            
            if player.JobEnum == JobEnum.Pet and self.Potency != 0: # Is a pet and action does damage

                # Updating damage and potency
                player.Master.TotalPotency+= self.Potency
                player.Master.TotalDamage += Damage
                player.Master.TotalMinDamage += minDamage

                # Updating Enemity
                if player.Master.RoleEnum == RoleEnum.Tank and player.Master.TankStanceOn:
                    # If the player is a tank and have their tank stance on
                    player.Master.TotalEnemity += Damage/1000
                    # This Enemity computation is arbitrary and is simply based on the fact that a tank with tank stance on
                    # generates 10 times the enemity of a player without tank stance.
                    # The value is made arbitrarily small in order to avoid too big numbers
                else:
                    player.Master.TotalEnemity += Damage/10000
            elif self.Potency != 0: # Is not a pet and action does damage
                # Updating damage and potency
                player.TotalPotency+= self.Potency
                player.TotalDamage += Damage
                player.TotalMinDamage += minDamage

                # Updating Enemity
                if player.RoleEnum == RoleEnum.Tank and player.TankStanceOn:
                    # If the player is a tank and have their tank stance on
                    player.TotalEnemity += Damage/1000
                    # This Enemity computation is arbitrary and is simply based on the fact that a tank with tank stance on
                    # generates 10 times the enemity of a player without tank stance.
                    # The value is made arbitrarily small in order to avoid too big numbers
                else:
                    player.TotalEnemity += Damage/10000

            Enemy.TotalPotency+= self.Potency  #Adding Potency
            Enemy.TotalDamage += Damage #Adding Damage

                                # This code starts the fight the first time a damaging action is done.
            if not (player.CurrentFight.FightStart) and Damage > 0 : 
                base_spell_logging.debug("Fight has started after the action "+name_for_id(player.CastingSpell.id,player.ClassAction, player.JobAction)+" done by player " + str(player.playerID))
                player.CurrentFight.FightStart = True
                                # Giving all players AA
                for gamer in player.CurrentFight.PlayerList:
                                # I feel like this could a place for pointer issue. Leaving this here so
                                # I see this in case there are issues and I forgor about that place.
                                # Note that caster/healer are not given AAs even though they in theory
                                # should. That will come in a future version.
                    aaDOT = None
                    #if gamer.JobEnum == JobEnum.Monk: aaDOT = copy.deepcopy(Monk_Auto)
                    if gamer.RoleEnum == RoleEnum.Melee or gamer.JobEnum == JobEnum.Dancer or gamer.RoleEnum == RoleEnum.Tank:
                        aaDOT = copy.deepcopy(Melee_AADOT)
                    elif gamer.RoleEnum == RoleEnum.PhysicalRanged:
                        aaDOT = copy.deepcopy(Ranged_AADOT)
                             # Giving AA
                    if aaDOT:
                        gamer.DOTList.append(aaDOT)
                        gamer.autoPointer = aaDOT
                                 # recomputing for current delay in case haste buffs were applied
                                 # before it starts. Only do this if the player has AA
                        gamer.recomputeRecastLock(isSpell=(gamer.RoleEnum == RoleEnum.Caster))

                                # Will record the starting HP of every player for graph
                for gamer in player.CurrentFight.PlayerList:
                    gamer.HPGraph[0].append(0)
                    gamer.HPGraph[1].append(gamer.HP)

                             # Will update the NextSpell of the player
        if (not (isinstance(self, DOTSpell))) : player.NextSpell+=1 # Only increase counter if action was not a DOT
                             # Checks if player has no more actions
        if (player.NextSpell == len(player.ActionSet)):
            if player.RoleEnum == RoleEnum.Pet: # If the player is a pet simply lock it
                player.TrueLock = True
            else: # Else we will call NextAction on this player before locking it
                player.NoMoreAction = True
            
        if self.GCD: player.GCDCounter += 1 # If action was a GCD, increase the counter

        if self.id > 0 or player.RoleEnum == RoleEnum.Pet or type==3: # Only logs if is a player action and not a DOT
            log_str = ( "Timestamp : " + str(player.CurrentFight.TimeStamp)
            + " , Event : end_cast"
            + (" , playerID : " + str(player.playerID) if player.JobEnum != JobEnum.Pet else " , MasterID : " + str(player.Master.playerID))
            + " , CastTime : " + str(self.CastTime) + " RecastTime : " + str(self.RecastTime)
            + " , Ability : " + name_for_id(self.id,player.ClassAction, player.JobAction)
            + " , SpellBonus : " + str(self.DPSBonus)
            + " , Potency : " + str(self.Potency)
            + (" , Damage : " + str(Damage) if not (self.AOEHeal or self.TargetHeal) else " , Healing : " + str(Heal)))
            
            base_spell_logging.debug(log_str)

        if self.Potency > 0: player.DamageInstanceList.append(Damage)
        if player.RoleEnum == RoleEnum.Pet and self.Potency > 0 : player.Master.DamageInstanceList.append(Damage)

        return self # Return the spell object. Might not be needed.

def ManaRequirement(player, Spell):
    """
    Requirement function for mana
    """
    if player.Mana >= Spell.ManaCost :
        player.Mana -= Spell.ManaCost   #ManaRequirement is the only Requirement that actually removes Ressources
        return True, -1
    return player.CurrentFight.IgnoreMana, -1 # Ignore mana is a field of the fight set to true if we ignore the mana

def empty(Player, Enemy):
    pass

def WaitAbility(time : float):
    """
    This returns an action where the player waits for a certain amount of time given
    time : float
    """
    def ApplyWaitAbility(Player, Enemy):
        pass
    WaitAction = Spell(212, False, time, time, 0, 0, ApplyWaitAbility, [])
    WaitAction.waitTime = time #Special field just for wait ability
    return WaitAction

def ApplyPotion(Player, Enemy):
    """
    Functions applies a potion and boosts the main stat of the player
    """
                             # The comp. % bonus is already applied on Player.Stat["MainStat"]
    Player.mainStatBonus = min(math.floor(Player.Stat["MainStat"] * 0.1),262) # Capped from grade 8 HQ tincture
    Player.Stat["MainStat"] += Player.mainStatBonus
    Player.PotionTimer = 30

    base_spell_logging.debug("Using Potion at " + str(Player.CurrentFight.TimeStamp))

    Player.EffectCDList.append(PotionCheck)

def PrepullPotion(Player, Enemy): #If potion is prepull
    """
    If the potion is prepull
    """
    ApplyPotion(Player, Enemy)
    Player.PotionTimer = 27 #Assume we loose a bit on it
    Player.EffectToRemove.append(PrepullPotion)

def PotionCheck(Player, Enemy):
    """
    Check of potion effect
    """
    if Player.PotionTimer <= 0:
        Player.Stat["MainStat"] -= Player.mainStatBonus #Assuming we are capped
        Player.EffectCDList.remove(PotionCheck)
        base_spell_logging.debug("removing Potion at " + str(Player.CurrentFight.TimeStamp))


class DOTSpell(Spell):
    """
    This class is any DOT. The action applying a dot will append a DOT object from this class (or any subclass of DOTSpell) which will do damage over time.
    """
    #Represents DOT
    def __init__(self, id, Potency, isPhysical, isGround : bool = False):
        """
        id : int -> id of the dot. Dot have negative ids
        Potency : int -> base potency of the DOT
        isPhysical : bool -> True if the dot is physical
        isGround : bool -> True if the DOT is ground. Ground DOT do not snapshot debuff.
        """
        super().__init__(id, False, 0, 0, Potency,  0, empty, [])
        #Note that here Potency is the potency of the dot, not of the ability
        self.DOTTimer = 0.02 # This represents the timer of the dot, and it will apply at each 3 seconds
                             # The timer is initialized at 0.02 because if it is initialized at 0 it hits immediatly, which
                             # results in 1 more tic than we would expect. This fixes the issue and gives the expected number of tic.
                             # Note that this could create issue in edge cases where buffs are not being clipped because there is less than
                             # 0.02 seconds left on them. I consider this negligeable for now.
        self.isPhysical = isPhysical #True if physical dot, false if magical dot
        self.isGround = isGround

        #This part will keep in memory the buffs when the DOT is applied.
        self.CritBonus = 0
        self.DHBonus = 0
        self.MultBonus = []
        self.potSnapshot = False
        self.onceThroughFlag = False #This flag will be set to True once the DOT damage has been through damage computation once
        self.ticAmount = 0 # Amount of tics done by this DOT object.
        #so we can snapshot the buffs only once
        #Note that AAs do not snapshot buffs, but in the code they will still have these fields

    def resetBuffSnapshot(self):
        """This function resets buff snapshots on the DOT object. used for when a DOT is reapplied before it falls off.
        """
        self.CritBonus = 0
        self.DHBonus = 0
        self.MultBonus = []
        self.potSnapshot = False
        self.onceThroughFlag = False

    def setBuffSnapshot(self, dotToCopy):
        """This function set the snapshots buff by copying the values in the given spellObj
        """
        self.CritBonus = dotToCopy.CritBonus
        self.DHBonus = dotToCopy.DHBonus
        self.MultBonus = dotToCopy.MultBonus
        self.potSnapshot = dotToCopy.potSnapshot

    def CheckDOT(self, Player, Enemy, TimeUnit : float):
        """
        This function is called every time unit of the simulation and will check if a dot will be applied. A dot is applied every 3 seconds.
        If a dot has to be applied it will Cast and Castfinal itself and reset its DOTTimer to 3 seconds.
        """

        self.DOTTimer = round(max(0, self.DOTTimer-TimeUnit),2)

        if(self.DOTTimer <= 0):
            #Apply DOT
            tempSpell  = self.Cast(Player, Enemy)#Cast the DOT
            tempSpell.CastFinal(Player, Enemy)
                             # If the DOT goes for the first time 
                             # it snapshots buff. We then transfer those
                             # buffs to the dot object that is being called.
                             # We will have to refresh those when the
                             # dot is reapplied.
            if not self.onceThroughFlag:
                self.onceThroughFlag = True
                # Using tempSpell to snapshot buffs
                self.setBuffSnapshot(tempSpell)

            self.ticAmount+=1 # A tic counter used for testing
                
            self.DOTTimer = 3
            
class HOTSpell(DOTSpell):
    """
    This represents a Healing Over Time effect.
    """

    def __init__(self, id, Potency):
        super().__init__(id, Potency, False)
                             # Every HOT is on only one target, hence they are targetted.
        self.TargetHeal = True


class Auto_Attack(DOTSpell):
    """
    DOTSpell subclass only for Autos since they have different potency depending on if ranged or melee.
    """
    def __init__(self, id, Ranged : bool):
        """
        Ranged : bool -> True if the auto is ranged.
        """
        if Ranged : super().__init__(id, 80, True) # Ranged AA, 80 potency
        else: super().__init__(id, 90, True) # Melee AA, 90 potency

        self.DOTTimer = 0 

    def CheckDOT(self, Player, Enemy, TimeUnit : float) -> None:
        """
        This is the function called to check if the AA is applied. Same function as to DOTSpell.CheckDOT
        """
                             # Update DOT Timer.
        self.DOTTimer = round(max(0, self.DOTTimer-TimeUnit),2)

        if(self.DOTTimer <= 0):
            #Apply DOT
            tempSpell  = self.Cast(Player, Enemy)#Cast the DOT
            tempSpell.CastFinal(Player, Enemy)
            
            self.DOTTimer = Player.currentDelay

        

class Queen_Auto(Auto_Attack):
    """
    Subclass of DOTSpell only for Machinist's queen autos
    """

    def __init__(self, id, Ranged):
        super().__init__(id, Ranged)
        self.Weaponskill = False

class Melee_Auto(Auto_Attack):
    """
    Subclass of DOTSpell only for melee autos
    """
    def __init__(self, id, Ranged):
        super().__init__(id, Ranged)
        self.Weaponskill = False

class Ranged_Auto(Auto_Attack):
    """
    Subclass of DOTSpell only for ranged autos
    """
    def __init__(self, id, Ranged):
        super().__init__(id, Ranged)
        self.Weaponskill = False

class Monk_AA(Melee_Auto):
    """
    Subclass of DOTSpell only for monk autos. The reason is that it can be on a faster rate if RiddleOfWind is activated. So the DOT
    update function is overwritten and checks for that and will update the timer accordingly.
    This class should in theory be deprecated and not used anymore.
    """
    def __init__(self):
        super().__init__(-5, False)
        self.DOTTimer = 0

    def CheckDOT(self, Player, Enemy, TimeUnit):
        
        self.DOTTimer = round(max(0, self.DOTTimer-TimeUnit),2)

        if(self.DOTTimer <= 0):
            #Apply AA
            tempSpell  = self.Cast(Player, Enemy)#Cast the DOT
            tempSpell.CastFinal(Player, Enemy)
            if Player.RiddleOfWindTimer > 0 : self.DOTTimer = 1.2
            else: self.DOTTimer = 2.4

def ApplyMonk_Auto(Player, Enemy):
    Player.DOTList.append(copy.deepcopy(Monk_Auto))

def ApplyMelee_AA(Player, Enemy):
    Player.DOTList.append(copy.deepcopy(Melee_AADOT))

def ApplyRanged_AA(Player, Enemy):
    Player.DOTList.append(copy.deepcopy(Ranged_AADOT))

def ApplyQueen_AA(Player, Enemy):
    Player.DOTList.append(copy.deepcopy(Queen_AADOT))

Melee_AA = Spell(-30, False, 0, 0, 0, 0, ApplyMelee_AA, [])
Ranged_AA = Spell(-30, False, 0, 0, 0, 0, ApplyRanged_AA, [])
Queen_AA = Spell(-30, False, 0, 0, 0, 0, ApplyQueen_AA, [])

Monk_Auto = Monk_AA()
Melee_AADOT = Melee_Auto(-22, False)
Ranged_AADOT = Ranged_Auto(-23, True)
Queen_AADOT = Queen_Auto(-24, False)
Potion = Spell(-2, False, 0.5, 0.5, 0, 0, ApplyPotion, [])

def conditionalAction(action : Spell, conditionFn) -> Spell:
    """
    This function returns an Spell object that will only be used if a certain condition is met. If
    the condition is not met the action is equivalent to WaitAbility(0).
    Note that the use of this will add 0.01 seconds to the end time if the condition evaluates to false
    action : Spell -> Action to perform if the condition evaluates to True
    conditionFn : function -> Function to evaluate. This function must return a boolean value. 
                              This function will be given the Player object as input when evaluated.
    """

    def __requirement(Player, Spell):
                             # The conditionFn will be evaluated in the requirement checking phase.
                             # If the conditionFn evaluates to false and we do not always allow conditional 
                             # action we effectively make this action WaitAbility(0).
        if not conditionFn(Player) and (not Player.CurrentFight.alwaysAllowConditionalAction) :
            Spell.CastTime = 0
            Spell.RecastTime = 0
            Spell.Potency = 0
            Spell.Effect = [] 
        return True, 0
    
    newAction = copy.deepcopy(action)
    newAction.Requirement.append(__requirement)
    newAction.conditionalAction = True
    return newAction

        

        






# This file will contain the enums of all jobs and class
from enum import IntEnum # Importing enums

class PlayerEnum(IntEnum):
    # Parent enum class for all other enums. Will have
    # the two functions.

    @classmethod
    def name_for_id(cls, id : int) -> str:
        # maps from id -> name
        if id in cls.__members__.values():
            return cls(id).name
        return 'Unknown'

    @classmethod
    def id_for_name(cls, name : str) -> int:
        # maps from name -> id
        if name in cls.__members__.keys():
            return cls[name].value
        return -1 # Evaluated as Unknown


class RoleEnum(PlayerEnum):
    # Enum for all roles

    Caster = 1
    Healer = 2
    Melee = 3
    Tank = 4
    PhysicalRanged = 5
    Pet = 6


class JobEnum(PlayerEnum):
    # Enum for all jobs

    # Caster
    BlackMage = 1
    Summoner = 2
    RedMage = 3

    # Healer
    WhiteMage = 4
    Astrologian = 5
    Sage = 6
    Scholar = 7

    # Melee
    Ninja = 8
    Samurai = 9
    Reaper = 10
    Monk = 11
    Dragoon = 12

    # Tank
    Gunbreaker = 13
    DarkKnight = 14
    Paladin = 15
    Warrior = 16

    # Physical ranged
    Machinist = 17
    Bard = 18
    Dancer = 19

    # Pet
    Pet = 20



#########################################
########## WHITEMAGE PLAYER #############
#########################################
from ffxivcalc.Jobs.Base_Spell import DOTSpell, empty, ManaRequirement
from ffxivcalc.Jobs.Healer.Healer_Spell import WhitemageSpell
import copy

#Requirement

def PresenceOfMindRequirement(Player, Spell):
    return Player.PresenceOfMindCD <= 0, Player.PresenceOfMindCD

def AssizeRequirement(Player, Spell):
    return Player.AssizeCD <= 0, Player.AssizeCD

def ThinAirRequirement(Player, Spell):
    return Player.ThinAirCD <= 0, Player.ThinAirCD

def BellRequirement(Player, Spell):
    return Player.BellCD <= 0, Player.BellCD

def AquaveilRequirement(Player, Spell):
    return Player.AquaveilCD <= 0, Player.AquaveilCD

def TemperanceRequirement(Player, Spell):
    return Player.TemperanceCD <= 0, Player.TemperanceCD

def PlenaryIndulgenceRequirement(Player, Spell):
    return Player.PlenaryIndulgenceCD <= 0, Player.PlenaryIndulgenceCD

def DivineBenisonRequirement(Player, Spell):
    return Player.DivineBenisonCD <= 0, Player.DivineBenisonCD

def TetragrammatonRequirement(Player, Spell):
    return Player.TetragrammatonCD <= 0, Player.TetragrammatonCD

def BenedictionRequirement(Player, Spell):
    return Player.BenedictionCD <= 0, Player.BenedictionCD

def AsylumRequirement(Player, Spell):
    return Player.AsylumCD <= 0, Player.AsylumCD

def BloodLilyRequirement(Player, Spell):
    return Player.LilyStack > 0, Player.LilyTimer

def BloomLilyRequirement(Player, Spell):
    return Player.BloomLily, -1



#Apply

def ApplyAfflatusMisery(Player, Enemy):
    Player.BloomLily = False
    Player.UsedLily = 0 #Reset counter

def ApplyLily(Player, Enemy):
    if not Player.BloomLily : Player.UsedLily = min(3, Player.UsedLily + 1)
    Player.LilyStack -= 1

    if Player.UsedLily == 3: Player.BloomLily = True

def Apply(Player, Enemy):
    Player.CD = 0

def ApplyBell(Player, Enemy):
    Player.BellCD = 180

def ApplyAquaveil(Player, Enemy):
    Player.AquaveilCD = 60

def ApplyTemperance(Player, Enemy):
    Player.TemperanceCD = 120

def ApplyPlenaryIndulgence(Player, Enemy):
    Player.PlenaryIndulgenceCD = 60

def ApplyDivineBenison(Player, Enemy):
    Player.DivineBenisonCD = 30

def ApplyTetragrammaton(Player, Enemy):
    Player.TetragrammatonCD = 60

def ApplyAsylum(Player, Enemy):
    Player.AsylumCD = 90

def ApplyBenediction(Player, Enemy):
    Player.BenedictionCD = 180

def ApplyDia(Player, Enemy):
    Player.DiaTimer = 30

    if (Player.Dia == None) : 
        Player.Dia = copy.deepcopy(DiaDOT)
        Player.EffectCDList.append(CheckDia)
        Player.DOTList.append(Player.Dia)
    else : Player.Dia.resetBuffSnapshot() # If already applied reset snapshot

def ApplyAssize(Player, Enemy):
    Player.Mana = min(10000, Player.Mana + 500)
    Player.AssizeCD = 40

def ApplyThinAir(Player, Enemy):
    Player.ThinAirCD = 60
    Player.EffectList.append(ThinAirEffect)

def ApplyPresenceOfMind(Player, Enemy):
    Player.PresenceOfMindCD = 120
    Player.PresenceOfMindTimer = 15
    Player.EffectList.append(PresenceOfMindEffect)
    Player.Haste += 20
    Player.hasteChangeValue = 20
    Player.hasteHasChanged = True
    Player.EffectCDList.append(CheckPresenceOfMind)

#Effect

def ThinAirEffect(Player, Spell):
    if Spell.GCD:
        Spell.ManaCost = 0
        Player.EffectList.remove(ThinAirEffect)

def PresenceOfMindEffect(Player, Spell):
    pass
    #Spell.CastTime *= 0.8
    #Spell.RecastTime *= 0.8

#Check

def CheckDia(Player, Enemy):
    if Player.DiaTimer <= 0:
        Player.DOTList.remove(Player.Dia)
        Player.Dia = None
        Player.EffectToRemove.append(CheckDia)

def CheckPresenceOfMind(Player, Enemy):
    if Player.PresenceOfMindTimer <= 0:
        Player.EffectList.remove(PresenceOfMindEffect)
        Player.Haste -= 20
        Player.hasteHasChanged = True
        Player.hasteChangeValue = -20
        Player.EffectToRemove.append(CheckPresenceOfMind)



#Damage GCD
Glare = WhitemageSpell(25859, True, 1.5, 2.5, 310, 400, empty, [ManaRequirement], type = 1)
Dia = WhitemageSpell(16532, True, 0, 2.5, 60, 400, ApplyDia, [ManaRequirement], type = 1)
DiaDOT = DOTSpell(-5, 60, False)
AfflatusMisery = WhitemageSpell(16535, True, 0, 2.5, 1240, 0, ApplyAfflatusMisery, [BloomLilyRequirement], type = 1)
Holy = WhitemageSpell(25860, True, 2.5, 2.5, 150, 400, empty, [ManaRequirement], type = 1)
#Healing GCD
AfflatusRapture = WhitemageSpell(16534, True, 0, 2.5, 0, 0, ApplyLily, [BloodLilyRequirement], type = 1, AOEHeal=True)
AfflatusSolace = WhitemageSpell(16531, True, 0, 0, 0, 0, ApplyLily, [BloodLilyRequirement], type = 1, TargetHeal=True)
Regen = WhitemageSpell(137, True, 0, 2.5, 0, 400, empty, [ManaRequirement], type = 1)
Cure = WhitemageSpell(120, True, 1.5, 2.5, 0, 400, empty, [ManaRequirement], type = 1, TargetHeal=True)
Cure2 = WhitemageSpell(135, True, 2, 2.5, 0, 1000, empty, [ManaRequirement], type = 1, TargetHeal=True)
Cure3 = WhitemageSpell(131, True, 2, 2.5, 0, 1500, empty, [ManaRequirement], type = 1, AOEHeal=True)
Medica = WhitemageSpell(124, True, 2, 2.5, 0, 900, empty, [ManaRequirement], type = 1, AOEHeal=True)
Medica2 = WhitemageSpell(133, True, 2, 2.5, 0, 1000, empty, [ManaRequirement], type = 1, AOEHeal=True)
Raise = WhitemageSpell(125, True, 8, 2.5, 0, 2400, empty, [ManaRequirement])

#Damage oGCD
Assize = WhitemageSpell(3571, False, 0, 0, 400, 0, ApplyAssize, [AssizeRequirement])
ThinAir = WhitemageSpell(7430, False, 0, 0, 0, 0, ApplyThinAir, [ThinAirRequirement])
PresenceOfMind = WhitemageSpell(136, False, 0, 0, 0, 0, ApplyPresenceOfMind, [PresenceOfMindRequirement])

#Healing oGCD
Bell = WhitemageSpell(25862, False, 0, 0, 0, 0, ApplyBell, [BellRequirement]) #Litturgy of the bell
Aquaveil = WhitemageSpell(25861, False, 0, 0, 0, 0, ApplyAquaveil, [AquaveilRequirement])
Temperance = WhitemageSpell(16536, False, 0, 0, 0, 0, ApplyTemperance, [TemperanceRequirement])
PlenaryIndulgence = WhitemageSpell(7433, False, 0, 0, 0, 0, ApplyPlenaryIndulgence, [PlenaryIndulgenceRequirement])
DivineBenison =WhitemageSpell(7432, False, 0, 0, 0, 0, ApplyDivineBenison, [DivineBenisonRequirement])
Tetragrammaton = WhitemageSpell(3570, False, 0, 0, 0, 0, ApplyTetragrammaton, [TetragrammatonRequirement], TargetHeal=True)
Asylum = WhitemageSpell(3569, False, 0, 0, 0, 0, ApplyAsylum, [AsylumRequirement])
Benediction = WhitemageSpell(140, False, 0, 0, 0, 0, ApplyBenediction, [BenedictionRequirement], TargetHeal=True)

WhiteMageAbility = {
25859 : Glare,
16532 : Dia,
25860 : Holy,
16535 : AfflatusMisery,
3571 : Assize,
120 : Cure,
135 : Cure2,
131 : Cure3,
124 : Medica,
133 : Medica2,
137 : Regen,
3570 : Tetragrammaton,
3569 : Asylum,
7432 : DivineBenison,
140 : Benediction,
16531 : AfflatusSolace,
16534 : AfflatusRapture,
136 : PresenceOfMind,
7430 : ThinAir,
16536 : Temperance,
25861 : Aquaveil,
25862 : Bell,
7433 : PlenaryIndulgence,
125 : Raise
}


"""
This code was written by Apollo (apollo.van.waddleburg on discord) and adapted to fit the needs of the sim.
"""


"""Defines and documents all custom columns in the data."""

# Not custom, but used a lot so recorded
ABILITY_ID = "abilityGameID"
TYPE = "type"
CAST = "cast"
DAMAGE = "damage"
APPLY_BUFF = "applybuff"
APPLY_DEBUFF = "applydebuff"
REFRESH_BUFF = "refreshbuff"
REMOVE_BUFF = "removebuff"
REMOVE_DEBUFF = "removedebuff"
PREPARE = "calculateddamage"
PLAYER = "Player"
PET = "Pet"

# Human readable name for abilities
ABILITY_NAME = "ability_name"


# Human readable source name for actors
SOURCE_NAME = "source_name"
SOURCE_TYPE = "source_type"
SOURCE_SUBTYPE = "source_subtype"
SOURCE_PET_ACTOR = "source_pet_actor"
SOURCE_ID = "sourceID"

# Human readable target name for actors
TARGET_NAME = "target_name"
TARGET_TYPE = "target_type"
TARGET_SUBTYPE = "target_subtype"
TARGET_PET_ACTOR = "target_pet_actor"
TARGET_ID = "targetID"

# Fight time, based on lb update tick to reset gauge
FIGHT_TIME = "fight_time"
SNAPSHOT = "snapshot"

# Bonus percent, estimate the value of combos
# 1 - basePotency/bonusPotency = bonusPercent
BONUS_PERCENT = "bonusPercent"

# Log hittype (crit, normal, etc)
LOG_HIT_TYPE = "hitType"
LOG_DHIT = "directHit"

# Special name added in for an assumed prepull apply
ESTIMATED_APPLY_BUFF = "estimatedapplybuff"

# Special name for "correct" card, other buffs
CARD_MATCHES_JOB = "card_matches_job"
RADIANT_CODA = "radiant_codas"

# Potency data and ability details
POTENCY = "ability_potency"
ABILITY_DAMAGE_TYPE = "ability_damage_type"  # See types below
DIRECT_DAMAGE = "direct_damage"
DOT = "dot"
AUTO = "auto"
GUARANTEED_CRIT = "guaranteed_crit"
GUARANTEED_DHIT = "guaranteed_dhit"
POSITIONAL = "positional"
SPECIAL_FORMULA = "special_formula"
GROUND_BASED = "ground_based"
ABILITY_TYPE = "ability_type"
GCD = "GCD"
OGCD = "OGCD"
LOCK_05 = "0.5 Lock"
LOCK_10 = "1.0 Lock"
LOCK_15 = "1.5 Lock"


# GCD and Weave tracking info
WEAVE = "weave"
GCD_ROLL = "gcd_roll"
NUM_WEAVES = "num_weaves"

# Buff related stuff
LOG_MULTIPLIER = "multiplier"
DAMAGE_MULTIPLIER = "damage_multiplier"
CRIT_BONUS = "crit_bonus"
DHIT_BONUS = "dhit_bonus"
MAIN_STAT_BONUS = "main_stat_bonus"
DOT_APPLICATION = "dot_application"
CAST_TIME = "cast_time"  # Defines the buff snapshot
LIMITED = "limited"  # a limited buff to specific types.
# Defines buff multipliers in a serial format as string
ACTIVE_BUFF_MULTIPLIERS = "active_buff_multipliers"
COMPLETE_DAMAGE_MULTIPLIER = "complete_damage_multiplier"
AVERAGED_MULTIPLIER = "averaged_multiplier"
CRIT_CHANCE = "crit_rate"
CRIT_MULT = "crit_multiplier"
CRIT_MOD = "crit_mod"
DHIT_CHANCE = "dhit_rate"
DHIT_MOD = "dhit_mod"
PERSONAL_DAMAGE_MULTIPLIER = "personal_damage_multiplier"  # This is all self buffs, reduce to get raw damage
RAW_DAMAGE_VARIANCE = "raw_damage_variance"  # variance without personal buffs
DAMAGE_VARIANCE = "damage_variance"  # variance with personal buffs

# Event types
HIT_TYPE = "hit_type"
CRIT = "crit"
DHIT = "direct_hit"
DCRIT = "direct_crit"
BASE = "base_hit"
AVERAGED = "averaged"
TICK = "tick"  # Helpful for determining tracking simulated dots
SIMULATED_DOT = "simulated_dot"

# Simmed Data
EXPECTED_DAMAGE = "expected_damage"
RAW_DAMAGE = "raw_damage"
BASE_DAMAGE = "base_damage"
AMOUNT = "amount"
LOG_DAMAGE = "unmitigatedAmount"  # amount column is affected by overkill
LOG_DOT_SIMULATED_DAMAGE = "finalizedAmount"  # Simmed dot damage
LATENT_VALUE = "latent_value"

# Damage taken types
DAMAGE_TYPE = "damage_type"
PHYSICAL = "physical"
MAGICAL = "magical"
DARKNESS = "darkness"
UNKNOWN = "unknown"

# Mitigation mods
PHYSICAL_MITIGATION = "Physical Mitigation"
MAGICAL_MITIGATION = "Magical Mitigation"
ADJUSTED_LOG_DAMAGE = (
    "unmitigatedBossAmount"  # amount column is affected by overkill
)

# Mappings (target_subtype)
JOB = "Job"

TARGET_SUBTYPE_TO_JOB_MELEE = {
    "DarkKnight": "DRK",
    "Gunbreaker": "GNB",
    "Paladin": "PLD",
    "Warrior": "WAR",
    "Ninja": "NIN",
    "Samurai": "SAM",
    "Reaper": "RPR",
    "Monk": "MNK",
    "Dragoon": "DRG",
}

TARGET_SUBTYPE_TO_JOB = {
    **TARGET_SUBTYPE_TO_JOB_MELEE,
    "Dancer": "DNC",
    "Bard": "BRD",
    "Machinist": "MCH",
    "RedMage": "RDM",
    "BlackMage": "BLM",
    "Summoner": "SMN",
    "Astrologian": "AST",
    "Scholar": "SCH",
    "Sage": "SGE",
    "WhiteMage": "WHM",
}

# ('type')
TYPE_TO_DAMAGE_TYPE = {
    "1": PHYSICAL,
    "32": DARKNESS,
    "64": MAGICAL,
    "128": PHYSICAL,
    "1024": MAGICAL,
}


"""
This file will contain basic function we will let the user call when importing the library or
functions callable by the API.
"""

from copy import deepcopy
import json
import math
from ffxivcalc.Jobs.PlayerEnum import JobEnum
from ffxivcalc.helperCode.helper_backend import GenerateLayoutBackend, GenerateLayoutDict, RestoreFightObject
from ffxivcalc.Fight import ComputeDamage
from ffxivcalc.Enemy import Enemy
from ffxivcalc.Jobs.Base_Spell import Spell
from ffxivcalc.Jobs.PlayerEnum import RoleEnum
from ffxivcalc.helperCode.helper_math import roundDown

# library

def AverageDamageAction(Player, Potency, MultBonus, type=0):

    levelMod = 1900
    baseMain = 390  
    baseSub = 400# Level 90 LevelMod values

    JobMod = Player.JobMod # Level 90 jobmod value, specific to each job

    Player.f_WD = (Player.Stat["WD"]+math.floor(baseMain*JobMod/1000))/100
    Player.f_DET = math.floor(1000+math.floor(140*(Player.Stat["Det"]-baseMain)/levelMod))/1000# Determination damage
    if Player.RoleEnum == RoleEnum.Tank : Player.f_TEN = (1000+math.floor(100*(Player.Stat["Ten"]-baseSub)/levelMod))/1000 # Tenacity damage, 1 for non-tank player
    else : Player.f_TEN = 1 # if non-tank
    Player.f_SPD = (1000+math.floor(130*((Player.Stat["SS"] if Player.RoleEnum == RoleEnum.Caster or Player.RoleEnum == RoleEnum.Healer else Player.Stat["SkS"])-baseSub)/levelMod))/1000 # Used only for dots
    Player.CritRate = math.floor((200*(Player.Stat["Crit"]-baseSub)/levelMod+50))/1000 # Crit rate in decimal
    Player.CritMult = (math.floor(200*(Player.Stat["Crit"]-baseSub)/levelMod+400))/1000 # Crit Damage multiplier
    Player.DHRate = math.floor(550*(Player.Stat["DH"]-baseSub)/levelMod)/1000 # DH rate in decimal

    return ComputeDamage(Player, Potency, Enemy(), MultBonus, type, Spell(0, False, 0, 0, 0, 0, None, []))


def CreateFightDict(PlayerList : list[JobEnum]) -> dict:
    """
    This function creates a fight dictionnary that can then be executed by the simulator. The fight's parameters will be default but
    it will create the fight with the given player list.
    PlayerList : list[JobEnum] -> Player list
    """
    return GenerateLayoutDict(PlayerList)

def CreateFightJSON(PlayerList : list[JobEnum], filename : str) -> None:
    """
    This function creates a dictionnary that the simulator can use and saves the file in a saved folder where the file is executed under the given name
    PlayerList : list[JobEnum] -> Player list
    filename : str -> name of the saved file
    """
    GenerateLayoutBackend(PlayerList, filename)

def AddPlayer(FightDict : dict, newPlayer : JobEnum) -> None:
    """
    This function will add a new player to the given Fight dictionnary. The new player will have no actions and have base stats.
    FightDict : dict -> Dictionnary with information of the fight
    newPlayer : JobEnum -> Job of the new player
    """
    jobName = JobEnum.name_for_id(newPlayer)
    if len(FightDict["data"]["PlayerList"]) == 0:
        new_id = 1
    else:
        new_id = FightDict["data"]["PlayerList"][-1]["playerID"] + 1
    # finding id of last player and adding 1 assuming they are in id order.


    newPlayerDict = {
        "JobName" : jobName,
        "playerID" : new_id,
        "stat": {
            "MainStat" : 390,
            "WD" : 0,
            "Det" : 390,
            "Ten" : 400,
            "SS" : 400,
            "Crit" : 400,
            "DH" : 400
        },
        "etro_gearset_url" : "put etro url here if you have one. Not needed",
        "Auras" : [],
        "actionLists" : [
            {
                "actionName" : "put action name here"
            }
        ]

    }

    FightDict["data"]["PlayerList"].append(newPlayerDict)

# function called by API

# GET

def GetSimulateFight(JSONFile):
    """
    This function will take the JSON file given with the API request and use other helper functions to Simulate the fight.
    JSONFile : json -> JSON file with the simulation's parameter
    """

    data = json.load(JSONFile) # Loading the file in a dictionnary

    returnData = SimulateFightAPIHelper(data)

    return json.dump(returnData,None, indent=4) # Returning the JSON file

def SimulateFightAPIHelper(FightDict : dict) -> dict:
    """
    This function can be called to simulate a fight. It requires as input the JSON file of the fight
    and it will return all the data relating to the simulation. This function will simply return the dictionnary and another function
    will return a JSON file. 
    FightDict : dict -> dictionnary holding the fight's info
    """

    Event = RestoreFightObject(FightDict) # Restoring the fight object
    fightInfo = FightDict["data"]["fightInfo"] #fight information
    Event.ShowGraph = fightInfo["ShowGraph"] #Default
    Event.RequirementOn = fightInfo["RequirementOn"]
    Event.IgnoreMana = fightInfo["IgnoreMana"]

    Event.SimulateFight(0.01,fightInfo["fightDuration"], vocal=False) # Simulating the fight

    # Now every player object has the info we need. We will simply parse is and put into a dictionnary. 
    # To return as a JSON file.

    # Skeleton of the returnData
    returnData = {
        "data" : {
            "fightInfo" : {
                "fightDuration" : Event.TimeStamp,
                "maxfightDuration" : fightInfo["fightDuration"],
                "fightname" : "ayo",
                "TeamCompositionBonus" : Event.TeamCompositionBonus,
                "failedRequirementEventList" : [],
                "Success" : True,
            },
            "PlayerList" : []
        }
    }

    # Will go through every failedRequirementEvent and record them
    success = True
    if FightDict["RequestParam"]["failedRequirementEvent"]: # If user wants failedRequirementList
        for event in Event.failedRequirementList:
            
            success = success and not event.fatal # if one event was fatal success is false

            eventDict = {
                "timeStamp" : event.timeStamp,
                "playerID" : event.playerID,
                "requirementName" : event.requirementName,
                "additionalInfo" : event.additionalInfo,
                "fatal" : event.fatal
            }

            returnData["data"]["fightInfo"]["failedRequirementEventList"].append(deepcopy(eventDict))

        returnData["data"]["fightInfo"]["Success"] = success


    for player in Event.PlayerList:
        # Going through every player will create the appriopriate dictionnary to return the info and put it in

        playerDict = {
            "JobName" : JobEnum.name_for_id(player.JobEnum),
            "ExpectedDPS" : 0 if Event.TimeStamp == 0 else round(player.TotalDamage/Event.TimeStamp,2),
            "PotencyPerSecond" :  0 if Event.TimeStamp == 0 else round(player.TotalPotency/Event.TimeStamp,2),
            "TotalDamage" : player.TotalDamage,
            "TotalPotency" : player.TotalPotency,
            "numberOfGCD" : player.GCDCounter,
            "ProcInfo" : {},
            "GraphInfoDPS" : [],
            "GraphInfoPPS" : []
        } # Creating new dictionnary

        # Going through its registered DPS and PPS points if user wants.

        if FightDict["RequestParam"]["GraphInfo"]:

            for (x,y) in zip(player.DPSGraph, Event.timeValue):
                # DPS

                point = {
                    "value" : y,
                    "name" : x
                }

                playerDict["GraphInfoDPS"].append(deepcopy(point))
            
            for (x,y) in zip(player.PotencyGraph, Event.timeValue):
                # PPS

                point = {
                    "value" : y,
                    "name" : x
                }

                playerDict["GraphInfoPPS"].append(deepcopy(point))

        # If any job has luck will return the info regarding this luck for the rotation
        if FightDict["RequestParam"]["ProcInfo"]: # If user wants
            match player.JobEnum:
                case JobEnum.Bard:
                    procInfo = {
                        "RefulgentArrow" : {
                            "Expected" : player.ExpectedRefulgent,
                            "Used" : player.UsedRefulgent
                        },
                        "WandererRepertoire" : {
                            "Expected" : player.ExpectedTotalWandererRepertoire,
                            "Used" : player.UsedTotalWandererRepertoire
                        },
                        "SoulVoiceGauge" : {
                            "Expected" : player.ExpectedSoulVoiceGauge,
                            "Used" : player.UsedSoulVoiceGauge
                        },
                        "BloodLetterReduction" : {
                            "Expected" : player.ExpectedBloodLetterReduction,
                            "Used" : player.UsedBloodLetterReduction
                        }
                    }
                    playerDict["ProcInfo"] = deepcopy(procInfo)
                case JobEnum.Dancer:
                    procInfo = {
                        "SilkenSymettry" : {
                            "Expected" : player.ExpectedSilkenSymettry,
                            "Used" : player.UsedSilkenSymettry
                        },
                        "SilkenFlow" : {
                            "Expected" : player.ExpectedSilkenFlow,
                            "Used" : player.UsedSilkenFlow
                        },
                        "FourfoldFeather" : {
                            "Expected" : player.ExpectedFourfoldFeather,
                            "Used" : player.UsedFourfoldFeather
                        },
                        "ThreefoldFan" : {
                            "Expected" : player.ExpectedThreefoldFan,
                            "Used" : player.UsedThreefoldFan
                        }
                    }
                    playerDict["ProcInfo"] = deepcopy(procInfo)
                case JobEnum.RedMage:
                    procInfo = {
                        "Verstone" : {
                            "Expected" : player.ExpectedVerstoneProc,
                            "Used" : player.UsedVerstoneProc
                        },
                        "Verfire" : {
                            "Expected" : player.ExpectedVerfireProc,
                            "Used" : player.UsedVerfireProc
                        }
                    }
                    playerDict["ProcInfo"] = deepcopy(procInfo)

        returnData["data"]["PlayerList"].append(deepcopy(playerDict))


    return returnData

"""
cur_dir = os.getcwd()

#the saved directory should be in that same folder

saved_dir = cur_dir + "\\saved"

saved_fight = os.listdir(saved_dir)


f = open(saved_dir + "\\" + "blackmage.json") #Opening save

data = json.load(f) #Loading json file

APIAnswer = SimulateFightAPIHelper(data)

with open(f'{"APIRequestTest"}.json', "w") as write_files:
    json.dump(APIAnswer,write_files, indent=4) #saving file

"""
# POST

    








import math

def isclose(a, b, rel_tol=1e-09, abs_tol=0.0): # Helper function to compare float
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)

def roundDown(x, precision):
    return math.floor(x * 10**precision)/10**precision

def roundUp(x, precision):
    return math.ceil(x * 10**precision)/10**precision


from ffxivcalc.Jobs.Base_Spell import WaitAbility
from ffxivcalc.Jobs.ActionEnum import name_for_id
#CASTER

from ffxivcalc.Jobs.Caster.Caster_Spell import CasterAbility
from ffxivcalc.Jobs.Caster.Blackmage.BlackMage_Spell import BlackMageAbility
from ffxivcalc.Jobs.Caster.Redmage.Redmage_Spell import RedMageAbility
from ffxivcalc.Jobs.Caster.Summoner.Summoner_Spell import SummonerAbility

#HEALER
from ffxivcalc.Jobs.Healer.Healer_Spell import HealerAbility
from ffxivcalc.Jobs.Healer.Sage.Sage_Spell import SageAbility
from ffxivcalc.Jobs.Healer.Astrologian.Astrologian_Spell import AstrologianAbility
from ffxivcalc.Jobs.Healer.Scholar.Scholar_Spell import ScholarAbility
from ffxivcalc.Jobs.Healer.Whitemage.Whitemage_Spell import WhiteMageAbility
from ffxivcalc.Jobs.Melee.Monk.Monk_Spell import MonkAbility

#RANGED
from ffxivcalc.Jobs.Ranged.Ranged_Spell import BardSpell, RangedAbility
from ffxivcalc.Jobs.Ranged.Bard.Bard_Spell import BardAbility
from ffxivcalc.Jobs.Ranged.Machinist.Machinist_Spell import MachinistAbility
from ffxivcalc.Jobs.Ranged.Dancer.Dancer_Spell import DancerAbility

#TANK
from ffxivcalc.Jobs.Tank.Tank_Spell import TankAbility
from ffxivcalc.Jobs.Tank.Gunbreaker.Gunbreaker_Spell import GunbreakerAbility
from ffxivcalc.Jobs.Tank.DarkKnight.DarkKnight_Spell import DarkKnightAbility
from ffxivcalc.Jobs.Tank.Warrior.Warrior_Spell import WarriorAbility
from ffxivcalc.Jobs.Tank.Paladin.Paladin_Spell import PaladinAbility

#MELEE
from ffxivcalc.Jobs.Melee.Melee_Spell import MeleeAbility
from ffxivcalc.Jobs.Melee.Samurai.Samurai_Spell import SamuraiAbility
from ffxivcalc.Jobs.Melee.Ninja.Ninja_Spell import NinjaAbility
from ffxivcalc.Jobs.Melee.Dragoon.Dragoon_Spell import DragoonAbility
from ffxivcalc.Jobs.Melee.Reaper.Reaper_Spell import ReaperAbility

from ffxivcalc.helperCode.exceptions import ActionNotFound, JobNotFound, InvalidTarget

import logging
main_logging = logging.getLogger("ffxivcalc")
lookUpAbilityId_logging = main_logging.getChild("lookUpAbilityId")

def lookup_abilityID(actionID, targetID, sourceID, player_list):
    """
    This function will translate an actionID into a Spell object of the relevant action that the simulator can use.
    actionID : int -> ID of the action in the game
    targetID : int -> ID of the target. Can be player or Enemy.
    sourceID : int -> ID of the player casting the action.
    player_list : dict -> dict of all players with some information.
    """
    #Will first get the job of the sourceID so we know in what dictionnary to search for

    def lookup(JobDict, ClassDict, job_name):
        """
        This function actually looks up the relevant dictionnary of the Job to find the Spell object.
        JobDict : dict -> dictionnary with keys being IDs and mapping to the Spell object (only for Job actions)
        ClassDict : dict -> same as JobDict, but for Class actions
        """
        if not (int(actionID) in JobDict.keys()): #if not in, then the action is in the ClassDict
            if not (int(actionID) in ClassDict.keys()):
                log_str = "Action Not found , Job : " + job_name + " , ActionId : " + str(actionID) + " , targetID : " + str(targetID) + " , sourceID : " + str(sourceID)
                lookUpAbilityId_logging.warning(log_str)
                lookUpAbilityId_logging.warning("Since action was not found defaulting to WaitAbility(0).")
                return WaitAbility(0) #Currently at none so we can debug
                raise ActionNotFound #Did not find action
            if callable(ClassDict[int(actionID)]): #If the action is a function
                if (not (str(targetID) in player_list.keys())):
                    player_obj = player_list[sourceID]["job_object"]
                    raise InvalidTarget(name_for_id(actionID, player_obj.ClassAction, player_obj.JobAction), player_obj, None,True, targetID)
                return ClassDict[int(actionID)](player_list[str(targetID)]["job_object"])
            return ClassDict[int(actionID)] #Class actions do not have the possibility to target other allies, so we assume itll target an enemy

        if callable(JobDict[int(actionID)]): #If the action is a function
            if (not (str(targetID) in player_list.keys())):
                player_obj = player_list[sourceID]["job_object"]
                raise InvalidTarget(name_for_id(actionID, player_obj.ClassAction, player_obj.JobAction), player_obj, None,True, targetID)
            return JobDict[int(actionID)](player_list[str(targetID)]["job_object"])
        return JobDict[int(actionID)] #Else return object

    job_name = player_list[str(sourceID)]["job"] #getting job name

    #Will now go through all possible job and find what action is being used based on the ID. If the ID is not right, it will
    #raise an ActionNotFoundError. And if the job's name does not exist it will raise a JobNotFoundError
    if job_name == "BlackMage" :#Caster
        return lookup(BlackMageAbility, CasterAbility,job_name)
    elif job_name == "RedMage":
        return lookup(RedMageAbility, CasterAbility,job_name)
    elif job_name == "Summoner":
        return lookup(SummonerAbility, CasterAbility,job_name)
    elif job_name == "Dancer":#Ranged
        return lookup(DancerAbility, RangedAbility,job_name)
    elif job_name == "Machinist":
        return lookup(MachinistAbility, RangedAbility,job_name)
    elif job_name == "Bard":
        return lookup(BardAbility, RangedAbility,job_name)
    elif job_name == "Warrior":#Tank
        return lookup(WarriorAbility, TankAbility,job_name)
    elif job_name == "Gunbreaker":
        return lookup(GunbreakerAbility, TankAbility,job_name)
    elif job_name == "DarkKnight":
        return lookup(DarkKnightAbility, TankAbility,job_name)
    elif job_name == "Paladin":
        return lookup(PaladinAbility, TankAbility,job_name)
    elif job_name == "WhiteMage":#Healer
        return lookup(WhiteMageAbility, HealerAbility,job_name)
    elif job_name == "Scholar":
        return lookup(ScholarAbility, HealerAbility,job_name)
    elif job_name == "Sage":
        return lookup(SageAbility, HealerAbility,job_name)
    elif job_name == "Astrologian":
        return lookup(AstrologianAbility, HealerAbility,job_name)
    elif job_name == "Samurai":#Melee
        return lookup(SamuraiAbility, MeleeAbility,job_name)
    elif job_name == "Reaper":
        return lookup(ReaperAbility, MeleeAbility,job_name)
    elif job_name == "Ninja":
        return lookup(NinjaAbility, MeleeAbility,job_name)
    elif job_name == "Monk":
        return lookup(MonkAbility, MeleeAbility,job_name)
    elif job_name == "Dragoon":
        return lookup(DragoonAbility, MeleeAbility,job_name)

    lookUpAbilityId_logging.critical("Job name not found : " + job_name)

    raise JobNotFound #If we get here, then we have not found the job in question
    #This should not happen, and if it does it means we either have a serious problem or the names aren't correct


"""
This code was written by Apollo (apollo.van.waddleburg on discord) and adapted to fit the needs of the sim.
"""

"""Defines views that integrate with ffxivcalc.

{
    "data": {
        "fightInfo": {
            "RequirementOn": true,
            "IgnoreMana": false,
            "fightDuration": "500"
        },
        "PlayerList": [
            {
                "JobName": "Samurai",
                "PlayerName": "SussyPlayer",
                "playerID": "0",
                "weaponDelay": "2.64",
                "stat": {
                    "MainStat": "3367",
                    "WD": "132",
                    "Det": "1736",
                    "Ten": "400",
                    "SS": "400",
                    "SkS": "400",
                    "Crit": "2587",
                    "DH": "1602"
                },
                "actionList": [
                ],
                "etro_gearset_url": "",
                "Auras": []
            }
        ]
    },
    "mode": false,
    "TeamCompBonus": true,
    "MaxPotencyPlentifulHarvest": false,
    "nRandomIterations": 0
}
"""
import functools
import logging
import os.path
from typing import Iterator, Dict, Union, Set, List

import pandas as pd
#import yaml
from ffxivcalc.Jobs import ActionEnum

from ffxivcalc.Request.etro_request import get_gearset_data
from ffxivcalc.GearSolver.Solver import getBaseStat
from ffxivcalc.Request import custom_columns
from ffxivcalc.Request import prepull

LOG = logging.getLogger(__name__)

_KEEP_TARGET_ACTION = [
    "Oblation",
    "TheBlackestNight",
    "HeartOfCorundum",
    "Bole",
    "Arrow",
    "Ewer",
    "Spear",
    "Spire",
    "Balance"
]

def ffxiv_sim_view(event_data: pd.DataFrame, max_time : float = 0) -> Iterator:
    casts_by_players_view = event_data[
        (event_data["type"] == "cast")
        & (event_data[custom_columns.SOURCE_TYPE] == "Player")
        | (event_data.index == event_data['abilityGameID'].dropna().index[0])
        | ((event_data.index == event_data[event_data['type'] == 'applybuff'].index[0]) & (event_data['fight_time'] != event_data[event_data['type'] == 'cast'].iloc[0]['fight_time']) & (event_data['fight_time'] < 0) )
    ]

    if max_time > 0:
        casts_by_players_view = casts_by_players_view[
            event_data["fight_time"] <= max_time
        ]
    
    # If first damage event then cast is not present and only calculated damage is.
    # The first 'applybuff' event is also not casted. So we add the first.

    casts_by_players_view = casts_by_players_view.astype(
        {
            custom_columns.SOURCE_ID: int,
            custom_columns.TARGET_ID: int,
            custom_columns.ABILITY_ID: int,
        }
    )
    fight_duration = (
        casts_by_players_view[custom_columns.FIGHT_TIME].max() + 10.0
    )
    player_ids = set(casts_by_players_view[custom_columns.SOURCE_ID].unique())
    player_list = []
    prepull_auras = prepull.prepull_aura_frame(fight_events=event_data)
    # prepull_estimated_casts = prepull.infer_prepull_buff_casts(fight_events=event_data)
    for (
        player_name,
        player_id,
        job_name,
    ), group in casts_by_players_view.groupby(
        by=[
            custom_columns.SOURCE_NAME,
            custom_columns.SOURCE_ID,
            custom_columns.SOURCE_SUBTYPE,
        ]
    ):
        actions = build_action_set(
            player_events=group,
            job_name=job_name,
            player_ids=player_ids,
        )

        # Removing target : 0
        for event in actions:
            if not (event['actionName'] in _KEEP_TARGET_ACTION) and (event['targetID'] == 0 or event['targetID'] == player_id):
                del event['targetID']

        # TODO: prepull_casts
        auras = build_aura_list(
            player_auras=prepull_auras[
                prepull_auras[custom_columns.SOURCE_NAME] == player_name
            ]
        )
        #etro_gearset_url = "url"#get_default_etro_url(job_name=job_name)
        #etro_stats = get_gearset_data(etro_gearset_url)

        etro_stats = getBaseStat(IsTank=False) #TODO make that work

        player_data = {
            "JobName": job_name,
            "PlayerName": player_name,
            "playerID": int(player_id),
            "weaponDelay": 3.0,#etro_stats.job.auto_speed,
            "stat": {
                "MainStat": etro_stats["MainStat"],
                "WD": etro_stats["WD"],
                "Det": etro_stats["Det"],
                "Ten": etro_stats["Ten"],
                "SS": etro_stats["SS"],
                "SkS": etro_stats["SkS"],
                "Crit": etro_stats["Crit"],
                "DH": etro_stats["DH"],
                "Piety" : etro_stats["Piety"]
            },
            "actionList": actions,
            "etro_gearset_url": "",
            "Auras": auras,
        }
        player_list.append(player_data)

    yield "ffxiv_sim.json", {
        "data": {
            "fightInfo": {
                "RequirementOn": True,
                "IgnoreMana": False,
                "fightDuration": fight_duration,
            },
            "PlayerList": player_list,
        },
        "mode": False,
        "TeamCompBonus": True,
        "MaxPotencyPlentifulHarvest": False,
        "nRandomIterations": 0,
    }


# Hard ignore on these events.
_IGNORE_IDS = {
    3, # Sprint
    7, # Autos
    8, # Autos
    33218, # Techbuff event
    27524, # Pneuma heal
    1000048 # Well fed
    }

_WAIT_OGCD = {
    "actionName": "WaitAbility",
    "waitTime": "0.65",
}

# Run replacements when these IDs are seen
_ID_REPLACEMENTS = {
    3: _WAIT_OGCD,  # Sprint
    3593: _WAIT_OGCD,
    16538: _WAIT_OGCD,  # Fey Illumn
    # 16558: "Horoscope",  # Activation?
    18873: "Ten",
    18877: "Chi",
    18881: "Jin",
}

# Run name replacements when these names are scene (other names have spaces removed by default)
_NAME_REPLACEMENTS = {
    "Medicated": "Potion",
    "Quadruple Technical Finish": "TechnicalFinish",
    "Double Standard Finish": "StandardFinish",
    "Braver": "MeleeLB1",
    "Bladedance": "MeleeLB2",
    "Chimatsuri": "MeleeLB3",
    **{
        lb_name: "MeleeLB3"
        for lb_name in [
            "Chimatsuri",
            "Final Heaven",
            "Doom of the Living",
            "Dragonsong Dive",
            "The End",
        ]
    },
}


def build_action_set(
    player_events,
    job_name: str,
    player_ids: Set[int],
) -> List[Dict]:
    """Builds an action set from a player's events."""
    x = 1
    return [
        {
            custom_columns.FIGHT_TIME: fight_time,
            custom_columns.ABILITY_ID: int(ability_id),
            **safe_check_ability_name( # This means it unpacks the returning dictionnary.
                ability_id=ability_id,
                ability_name=ability_name,
                job_name=job_name,
            ),
            # TargetID = 0 implies boss.
            custom_columns.TARGET_ID: (
                target_id if (target_id in player_ids) else 0
            ),
        }
        for fight_time, ability_id, ability_name, target_id in zip(
            player_events[custom_columns.FIGHT_TIME],
            player_events[custom_columns.ABILITY_ID],
            player_events[custom_columns.ABILITY_NAME],
            player_events[custom_columns.TARGET_ID],
        )
        if int(ability_id) not in _IGNORE_IDS  # Ignore autos, double tech
    ]


def safe_check_ability_name(
    ability_id: int, ability_name: str, job_name: str
) -> Dict[str, str]:
    """Handles conversion from ability ID to ffxiv-calc name + params."""
    output = _get_ffxiv_calc_ability_name(
        ability_id=ability_id,
        ability_name=ability_name,
        job_name=job_name,
    )
    return (
        output
        if isinstance(output, dict)
        else {
            "actionName": output,
        }
    )

_NAME_IGNORE = [
    "Well fed"
]


@functools.cache
def _get_ffxiv_calc_ability_name(
    ability_id: int, ability_name: str, job_name: str
) -> Union[Dict[str, str], str]:
    """Inner helper to handle all the edge-cases."""
    if ability_name in _NAME_IGNORE:
        return None
    if ability_name in _NAME_REPLACEMENTS:
        return _NAME_REPLACEMENTS[ability_name]
    if ability_id in _ID_REPLACEMENTS:
        return _ID_REPLACEMENTS[ability_id]
    cls_enum, job_enum = SOURCE_SUBTYPE_TO_ENUMS[job_name]
    name_from_id = ActionEnum.name_for_id(
        id=ability_id,
        cls=cls_enum,
        job_cls=job_enum,
    )

    if name_from_id != "Unknown":
        return name_from_id

    # Check name (with no whitespace) isn't mapped to another id
    ability_name_trimmed = ability_name.replace(" ", "")
    id_from_name = ActionEnum.id_for_name(
        name=ability_name_trimmed,
        cls=cls_enum,
        job_cls=job_enum,
    )

    if id_from_name != -1:
        LOG.debug(
            f"Ability {ability_id} corrected to {id_from_name} ({ability_name_trimmed})"
        )
        return ability_name_trimmed

    LOG.debug(f"Unmatched ID {ability_id} ({ability_name}): Assuming wait OGCD")
    return _WAIT_OGCD


AURA_MAPPINGS = {
    "Sharpcast": "SharpCast",
    "Ley Lines": "Ley Lines",
    "Soulsow": "Soulsow",
    "Medicated": "Medicated",
    "Meikyo Shisui": "Meikyo Shisui",
    "Mudra": "Mudra",
    "Eukrasia": "Eukrasia",
    "Standard Step": "Standard Step",
    "Standard Finish": "Standard Finish",
    "Closed Position": "Closed Position",
    "Dance Partner": "Dance Partner",
    "Technical Step" : "Technical Step",
    "Triplecast" : "Triplecast",
    "Inner Release" : "Inner Release",
    "Fight or Flight" : "Fight or Flight",
    "Delirium" : "Delirium",
    "Blood Weapon" : "Blood Weapon",
    "Blackest Night" : "Blackest Night",
    "No Mercy" : "No Mercy",
    "Arcane Circle" : "Arcane Circle",
    "Riddle of Fire" : "Riddle of Fire",
    "Riddle of Wind" : "Riddle of Wind",
    "Life Surge" : "Life Surge",
    "Lance Charge" : "Lance Charge",
    "Battle Litany" : "Battle Litany",
    "Right Eye" : "Right Eye",
    "Left Eye" : "Left Eye",
    "Kassatsu" : "Kassatsu",
    "Battle Voice" : "Battle Voice",
    "Acceleration" : "Acceleration",
    "Embolden" : "Embolden",
    "Searing Light" : "Searing Light",
    "Presence of Mind" : "Presence of Mind",
    "Earthly Dominance" : "Earthly Dominance",
    "Arrow Drawn": "Arrow Drawn",
    "Balance Drawn": "Balance Drawn",
    "Ewer Drawn": "Ewer Drawn",
    "Spire Drawn": "Spire Drawn",
    "Spear Drawn": "Spear Drawn",
    "Bole Drawn": "Bole Drawn",
    "Royal Guard" : "Royal Guard",
    "Grit" : "Grit",
    "Iron Will" : "Iron Will",
    "Defiance" : "Defiance",
    "Reassembled" : "Reassembled"
}


def build_aura_list(player_auras: pd.DataFrame) -> List[str]:
    """Builds a list of pre-defined mapped auras."""
    if player_auras.empty:
        return []
    return [
        AURA_MAPPINGS[aura]
        for aura, job_name in zip(
            player_auras["name"], player_auras["source_subtype"]
        )
        if aura in AURA_MAPPINGS
        and not (
            # Ignore standard finish if job is not dancer
            aura == "Standard Finish"
            and job_name != "Dancer"
        )
    ]


SOURCE_SUBTYPE_TO_ENUMS = {
    "DarkKnight": (ActionEnum.TankActions, ActionEnum.DarkKnightActions),
    "Gunbreaker": (ActionEnum.TankActions, ActionEnum.GunbreakerActions),
    "Paladin": (ActionEnum.TankActions, ActionEnum.PaladinActions),
    "Warrior": (ActionEnum.TankActions, ActionEnum.WarriorActions),
    "Ninja": (ActionEnum.MeleeActions, ActionEnum.NinjaActions),
    "Samurai": (ActionEnum.MeleeActions, ActionEnum.SamuraiActions),
    "Reaper": (ActionEnum.MeleeActions, ActionEnum.ReaperActions),
    "Monk": (ActionEnum.MeleeActions, ActionEnum.MonkActions),
    "Dragoon": (ActionEnum.MeleeActions, ActionEnum.DragoonActions),
    "Dancer": (ActionEnum.RangedActions, ActionEnum.DancerActions),
    "Bard": (ActionEnum.RangedActions, ActionEnum.BardActions),
    "Machinist": (ActionEnum.RangedActions, ActionEnum.MachinistActions),
    "RedMage": (ActionEnum.CasterActions, ActionEnum.RedMageActions),
    "BlackMage": (ActionEnum.CasterActions, ActionEnum.BlackMageActions),
    "Summoner": (ActionEnum.CasterActions, ActionEnum.SummonerActions),
    "Astrologian": (ActionEnum.HealerActions, ActionEnum.AstrologianActions),
    "Scholar": (ActionEnum.HealerActions, ActionEnum.ScholarActions),
    "Sage": (ActionEnum.HealerActions, ActionEnum.SageActions),
    "WhiteMage": (ActionEnum.HealerActions, ActionEnum.WhiteMageActions),
}


#@functools.cache
#def get_yaml_schema(file_name: str) -> Dict:
#    with open(file_name, "r") as file:
#        return yaml.safe_load(file)


#def get_default_etro_url(job_name: str) -> str:
#    """Inner helper to handle all the edge-cases."""
#    return get_yaml_schema(
#        os.path.join("main", "calc", "comps", "default.yaml")
#    )["6.4"][custom_columns.TARGET_SUBTYPE_TO_JOB[job_name]]


import copy
import json
import logging
main_logging = logging.getLogger("ffxivcalc")
helper_logging = main_logging.getChild("Helper")
from pathlib import Path
import math

from ffxivcalc.Jobs.Melee.Monk.Monk_Spell import ComboEffect
from ffxivcalc.Jobs.Melee.Ninja.Ninja_Spell import Huton, Kassatsu
from ffxivcalc.Jobs.Melee.Samurai.Samurai_Spell import Meikyo
from ffxivcalc.Jobs.Ranged.Dancer.Dancer_Spell import EspritEffect, ClosedPosition, StandardStep, Pirouette, Emboite, Entrechat, Jete, TechnicalStep
from ffxivcalc.Jobs.Ranged.Bard.Bard_Spell import SongEffect, BattleVoice
from ffxivcalc.Jobs.Ranged.Machinist.Machinist_Spell import Reassemble
from ffxivcalc.Jobs.Tank.DarkKnight.DarkKnight_Spell import BloodWeaponCheck, BloodWeaponEffect, Delirium, BloodWeapon, TBN
from ffxivcalc.Jobs.Tank.Warrior.Warrior_Spell import InnerRelease
from ffxivcalc.Jobs.Caster.Redmage.Redmage_Spell import DualCastEffect
from ffxivcalc.Jobs.Caster.Blackmage.BlackMage_Spell import ElementalEffect, SharpCast, LeyLines, Triplecast
from ffxivcalc.Jobs.Healer.Sage.Sage_Spell import Eukrasia
from ffxivcalc.Jobs.Healer.Astrologian.Astrologian_Spell import EarthlyStar, Draw
from ffxivcalc.Jobs.Tank.Paladin.Paladin_Spell import FightOrFlight
from ffxivcalc.Jobs.Tank.Gunbreaker.Gunbreaker_Spell import NoMercy
from ffxivcalc.Jobs.Melee.Reaper.Reaper_Spell import ArcaneCircle, Soulsow
from ffxivcalc.Jobs.Melee.Monk.Monk_Spell import RiddleOfFire, RiddleOfWind, Brotherhood
from ffxivcalc.Jobs.Melee.Dragoon.Dragoon_Spell import LifeSurge, LanceCharge, BattleLitany, DragonSight
from ffxivcalc.Jobs.Caster.Redmage.Redmage_Spell import Acceleration, Embolden
from ffxivcalc.Jobs.Caster.Summoner.Summoner_Spell import SearingLight
from ffxivcalc.Jobs.Healer.Whitemage.Whitemage_Spell import PresenceOfMind
from ffxivcalc.Jobs.Tank.Tank_Spell import Defiance, Grit, IronWill, RoyalGuard

from ffxivcalc.Fight import Fight
from ffxivcalc.Enemy import Enemy
from ffxivcalc.helperCode.lookUpAbilityID import lookup_abilityID
from ffxivcalc.Jobs.Base_Spell import Potion, WaitAbility
from ffxivcalc.Jobs.Player import Player
from ffxivcalc.Jobs.PlayerEnum import *

from ffxivcalc.Jobs.ActionEnum import name_for_id, id_for_name # importing helper functions
from ffxivcalc.Request.etro_request import get_gearset_data


letters = "abcdefghijklmnopqrstuvwyxz" # Used to make sure the input is only numbers

#This file will take care of request by TUI.py and return whatever needs to be returned

# ============================================================================================

# These are the stats that will be used when importing from FFLogs. Make sure to change them accordingly

# Caster
BLMStat = {"MainStat": 2945, "WD":126, "Det" : 1451, "Ten" : 400, "SS": 840, "SkS" : 400,  "Crit" : 2386, "DH" : 1307} # Stats for BlackMage
RDMStat = {"MainStat": 2947, "WD":126, "Det" : 1548, "Ten" : 400, "SS": 495, "SkS" : 400, "Crit" : 2397, "DH" : 1544} # Stats for RedMage
SMNStat = {"MainStat": 2948, "WD":126, "Det" : 1451, "Ten" : 400, "SS": 544, "SkS" : 400, "Crit" : 2436, "DH" : 1544} # Stats for Summoner

# Healer
SCHStat = {"MainStat": 2931, "WD":126, "Det" : 1750, "Ten" : 400, "SS": 1473, "SkS" : 400, "Crit" : 2351, "DH" : 436} # Stats for Scholar
WHMStat = {"MainStat": 2945, "WD":126, "Det" : 1792, "Ten" : 400, "SS": 839, "SkS" : 400, "Crit" : 2313, "DH" : 904} # Stats for WhiteMage
ASTStat = {"MainStat": 2949, "WD":126, "Det" : 1659, "Ten" : 400, "SS": 1473, "SkS" : 400, "Crit" : 2280, "DH" : 436} # Stats for Astrologian
SGEStat = {"MainStat": 2928, "WD":126, "Det" : 1859, "Ten" : 400, "SS": 827, "SkS" : 400, "Crit" : 2312, "DH" : 1012} # Stats for Sage

# Physical Ranged
MCHStat = {"MainStat": 2937, "WD":126, "Det" : 1598, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2389, "DH" : 1592} # Stats for Machinist
BRDStat = {"MainStat": 2949, "WD":126, "Det" : 1721, "Ten" : 400, "SS": 400, "SkS" : 536, "Crit" : 2387, "DH" : 1340} # Stats for Bard
DNCStat = {"MainStat": 2949, "WD":126, "Det" : 1721, "Ten" : 400, "SS": 400, "SkS" : 536, "Crit" : 2387, "DH" : 1340} # Stats for Dancer

# Melee
NINStat = {"MainStat": 2921, "WD":126, "Det" : 1669, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2399, "DH" : 1511} # Stats for Ninja
SAMStat = {"MainStat": 2937, "WD":126, "Det" : 1571, "Ten" : 400, "SS": 400, "SkS" : 508, "Crit" : 2446, "DH" : 1459} # Stats for Samurai
DRGStat = {"MainStat": 2949, "WD":126, "Det" : 1545, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2462, "DH" : 1577} # Stats for Dragoon
MNKStat = {"MainStat": 3076, "WD":126, "Det" : 1546, "Ten" : 400, "SS": 400, "SkS" : 769, "Crit" : 2490, "DH" : 1179} # Stats for Monk
RPRStat = {"MainStat": 2946, "WD":126, "Det" : 1545, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2462, "DH" : 1577} # Stats for Reaper

# Tank
DRKStat = {"MainStat": 2910, "WD":126, "Det" : 1844, "Ten" : 751, "SS": 400, "SkS" : 400, "Crit" : 2377, "DH" : 1012} # Stats for DarkKnight
WARStat = {"MainStat": 2910, "WD":126, "Det" : 1844, "Ten" : 751, "SS": 400, "SkS" : 400, "Crit" : 2377, "DH" : 1012} # Stats for Warrior
PLDStat = {"MainStat": 2891, "WD":126, "Det" : 1883, "Ten" : 631, "SS": 400, "SkS" : 650, "Crit" : 2352, "DH" : 868} # Stats for Paladin
GNBStat = {"MainStat": 2891, "WD":126, "Det" : 1883, "Ten" : 631, "SS": 400, "SkS" : 650, "Crit" : 2352, "DH" : 868} # Stats for Gunbreaker

# ============================================================================================

def ImportFightBackend(fightID,fightNumber):
    """This function is to no longer be used."""
    return 
    Event = Fight(Enemy(), False) #Creating event
    helper_logging.debug("Constructing Event object from FFLogs link.")
    action_dict, player_dict = getAbilityList(fightID, fightNumber) #getting ability List
    
    for playerID in player_dict:

        player_dict[playerID]["job_object"].ActionSet = action_dict[playerID]
        player_dict[playerID]["job_object"].CurrentFight = Event
        job_name = player_dict[playerID]["job"] #getting job name

        if job_name == "Sage" : player_dict[playerID]["job_object"].Stat = SGEStat
        elif job_name == "Scholar" : player_dict[playerID]["job_object"].Stat = SCHStat
        elif job_name == "WhiteMage" : player_dict[playerID]["job_object"].Stat = WHMStat
        elif job_name == "Astrologian" : player_dict[playerID]["job_object"].Stat = ASTStat
        elif job_name == "Warrior" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(WARStat)
        elif job_name == "DarkKnight" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(DRKStat)
            player_dict[playerID]["job_object"].EffectList = [BloodWeaponEffect] #Assuming we pre pull it
            player_dict[playerID]["job_object"].EffectCDList = [BloodWeaponCheck]
        elif job_name == "Paladin" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(PLDStat)
        elif job_name == "Gunbreaker" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(GNBStat)
        #Caster
        elif job_name == "BlackMage" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(BLMStat)
        elif job_name == "RedMage" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(RDMStat)
            player_dict[playerID]["job_object"].EffectList = [DualCastEffect]
        elif job_name == "Summoner" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(SMNStat)
        #Ranged
        elif job_name == "Dancer" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(DNCStat)
            player_dict[playerID]["job_object"].EffectList = [EspritEffect]
        elif job_name == "Machinist" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(MCHStat)
        elif job_name == "Bard" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(BRDStat)
            player_dict[playerID]["job_object"].EffectList = [SongEffect]
        #melee
        elif job_name == "Reaper" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(RPRStat)
        elif job_name == "Monk" :              
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(MNKStat)

        elif job_name == "Dragoon" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(DRGStat)
        elif job_name == "Ninja" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(NINStat)
        elif job_name == "Samurai" : 
            player_dict[playerID]["job_object"].Stat = copy.deepcopy(SAMStat)

        Event.PlayerList.append(player_dict[playerID]["job_object"])

    Event.RequirementOn = False #By default making false

    return Event

def AskInput(range):
    #This function will ask a numerical value to the user where range is the number of options
    #The inputs are assumed to be numerically ordered and starting at 1 and finishing at (range)
    user_input = input("Enter a number to select that option : ")

    while True:

        has_letter = False

        for letter in user_input:
            if letter.lower() in letters: has_letter = True

        if user_input == "" or has_letter: 
            has_letter = False
            user_input = input("This is not a valid input. Please enter a valid number : ")
        elif int(user_input) >= 1 and int(user_input) <= range: #if input is valid
            return user_input
        else : #Non-valid input
            user_input = input("This is not a valid input. Please enter a valid number : ")

def SaveFight(Event, countdown, fightDuration, saveName, saveFile=True):
    #This function will save a fight into memory.

    #The file will be saved as a JSON format

    PlayerListDict = []
    PlayerIDList = [] #list of all player IDs so we don't have duplicates

    for Player in Event.PlayerList:
        #Going through all players in the Event
        PlayerDict = {} #Empty dictionnary


        PlayerDict["JobName"] = JobEnum.name_for_id(Player.JobEnum)

        while Player.playerID in PlayerIDList:
            Player.playerID += 1

        PlayerDict["playerID"] = Player.playerID

        PlayerIDList += [Player.playerID]

        PlayerDict["stat"] = Player.Stat
                             # Using base main stat since it will have been affected if the fight has already been simulated
        PlayerDict["stat"]["MainStat"] = Player.baseMainStat
        actionList = []

        PlayerDict["Auras"] = copy.deepcopy(Player.auras)

        for action in Player.ActionSet:
            actionDict = {}
            #Special case for WaitAbility
            if action.id == 212 : #WaitAbility
                if action.waitTime != 0:
                    actionDict["actionName"] = name_for_id(action.id,Player.ClassAction, Player.JobAction)
                    actionDict["waitTime"] = action.waitTime

                    actionList.append(copy.deepcopy(actionDict))#adding to dict
            else: #Normal ability
                actionDict["actionName"] = name_for_id(action.id,Player.ClassAction, Player.JobAction)
                if action.TargetID != 0 : 
                    actionDict["targetID"] = action.TargetPlayerObject.playerID # id 0 is by default the main enemy. So no need to write if not needed

                actionList.append(copy.deepcopy(actionDict))#adding to dict

        PlayerDict["actionList"] = copy.deepcopy(actionList)
        PlayerDict["etro_gearset_url"] = "Put a URL here if you want. The code will only overwirte the stats if it detects an etro url."

        PlayerListDict.append(copy.deepcopy(PlayerDict))

    


    data = {"data" : {
                "fightInfo" : {
                    "countdownValue" : countdown,
                    "fightDuration" : fightDuration,
                    "time_unit" : 0.01,
                    "ShowGraph" : Event.ShowGraph,
                    "RequirementOn" : Event.RequirementOn,
                    "IgnoreMana" : Event.IgnoreMana

                },
                "PlayerList" : PlayerListDict
    }}

    if saveFile:
                             # If saving as file save it
        save_dir: Path = Path.cwd() / 'saved'
        with open(save_dir / f'{saveName}.json', "w") as write_files:
            json.dump(data,write_files, indent=4) #saving file
    return data

def RestoreFightObject(data : dict, name : str = ""):
    """
    This takes a FightDict dictionnary and converts it back into an Event object.
    data : dict -> dictionnary with the fight's data
    """

    helper_logging.debug("Restoring saved file " + name + " into Event object.")

    
    PlayerActionList = {} #Dictionnary containing all player with their action

    foundDancePartner = False
    foundClosedPosition = False
    dancerId = None
    partnerId = None

    foundRightEye = False
    foundLeftEye = False
    leftEyeTargetId = -1
    rightEyeId = -1



    for player in data["data"]["PlayerList"]: #Going through all player in PlayerList and creating JobObject
        #Will check what job the player is so we can create a player object of the relevant job
        
        job_name = player["JobName"]
        job_object = None
        stat = {}
        #Giving player object the stat dictionnary
        
        # Will check if the player is given an etro url.

        etro_url = player["etro_gearset_url"]
        if "etro.gg/gearset/" in etro_url:
            stat = get_gearset_data(etro_url)

            # Since etro now returns only the weapon damage multiplier, we have to find the the original WD.
            #job_object.f_WD = PlayerActionList[str(job_object.playerID)]["job_object"].Stat["WD"]/100 # Mult bonus we already know.
            
            #baseMain = 390
            #JobMod = job_object.JobMod # Level 90 jobmod value, specific to each job
            #stat["WD"] -= math.floor((baseMain * JobMod) /1000)
        else :
            stat = player["stat"]
        
        for key in stat:
            stat[key] = int(stat[key])

        #Healer
        if job_name == "Sage" : job_object = Player([], [],stat, JobEnum.Sage)
        elif job_name == "Scholar" : job_object = Player([], [], stat, JobEnum.Scholar)
        elif job_name == "WhiteMage" : job_object = Player([], [], stat, JobEnum.WhiteMage)
        elif job_name == "Astrologian" : job_object = Player([], [], stat, JobEnum.Astrologian)
        #Tank
        elif job_name == "Warrior" : job_object = Player([], [],  stat, JobEnum.Warrior)
        elif job_name == "DarkKnight" : job_object = Player([], [],  stat, JobEnum.DarkKnight)
        elif job_name == "Paladin" : job_object = Player([], [], stat, JobEnum.Paladin)
        elif job_name == "Gunbreaker" : job_object = Player([], [],  stat, JobEnum.Gunbreaker)
        #Caster
        elif job_name == "BlackMage" : job_object = Player([], [ElementalEffect],  stat, JobEnum.BlackMage)
        elif job_name == "RedMage" : job_object = Player([], [DualCastEffect],  stat, JobEnum.RedMage)
        elif job_name == "Summoner" : job_object = Player([], [],  stat, JobEnum.Summoner)
        #Ranged
        elif job_name == "Dancer" : job_object = Player([], [EspritEffect],  stat, JobEnum.Dancer)
        elif job_name == "Machinist" : job_object = Player([], [],  stat, JobEnum.Machinist)
        elif job_name == "Bard" : job_object = Player([], [SongEffect],  stat, JobEnum.Bard)
        #melee
        elif job_name == "Reaper" : job_object = Player([], [],  stat, JobEnum.Reaper)
        elif job_name == "Monk" : job_object = Player([], [ComboEffect],  stat, JobEnum.Monk)
        elif job_name == "Dragoon" : job_object = Player([], [],  stat, JobEnum.Dragoon)
        elif job_name == "Ninja" : job_object = Player([], [],  stat, JobEnum.Ninja)
        elif job_name == "Samurai" : job_object = Player([], [],  stat, JobEnum.Samurai)
        

        job_object.playerID = player["playerID"] #Giving the playerID
                             # If weaponDelay was in the data we set it
        if 'weaponDelay' in player.keys() : job_object.setBasedWeaponDelay(float(player['weaponDelay']))
        helper_logging.debug("Creating job object : " + job_name + " for playerID : " + str(player["playerID"]))

        if "PlayerName" in player.keys():job_object.PlayerName = player["PlayerName"]

        PlayerActionList[str(job_object.playerID)] = {"job" : job_name, "job_object" : job_object, "actionList" : player["actionList"], "actionObject" : []} #Adding new Key accessible by IDs

        # If detects an etro url copies the stats. Otherwise takes the stats from the file
        #We can access the information using the player's id
        #We will then check for Auras and do the appropriate effect
        
        # Logging the stat of the player
        helper_logging.debug("Player id " + str(job_object.playerID) + " has stats : " + str(job_object.Stat))

        for aura in player["Auras"]: #Going through all buffs in the player
            #We will look for a selection of buffs that are important. We will assume
            #the optimal scenario. So if we use a potion, we will assume its right before the fight begins
            
            job_object.auras += [aura] #Adding aura. Used for when restoring a fight using a saved file.
            
            if aura == "SharpCast":
                job_object.ActionSet.append(SharpCast)
            elif aura == "Ley Lines":
                job_object.ActionSet.append(LeyLines)
            elif aura == "Triplecast":
                job_object.ActionSet.append(Triplecast)
            elif aura == "Medicated":
                job_object.ActionSet.append(Potion)
            elif aura == "Inner Release":
                job_object.ActionSet.append(InnerRelease)
            elif aura == "Fight or Flight":
                job_object.ActionSet.append(FightOrFlight)
            elif aura == "Delirium":
                job_object.ActionSet.append(Delirium)
            elif aura == "Blood Weapon":
                job_object.ActionSet.append(BloodWeapon)
            elif aura == "Blackest Night":
                if job_object.JobEnum == JobEnum.DarkKnight:
                    job_object.ActionSet.append(TBN(job_object))
                    # TODO Add code that makes the darkknight cast it. Now only casts if self cast from DRK
            elif aura == "No Mercy":
                job_object.ActionSet.append(NoMercy)
            elif aura == "Arcane Circle":
                if job_object.JobEnum == JobEnum.Reaper:
                    job_object.ActionSet.append(ArcaneCircle)
            elif aura == "Soulsow":
                job_object.ActionSet.append(Soulsow)
            elif aura == "Riddle of Fire":
                job_object.ActionSet.append(RiddleOfFire)
            elif aura == "Riddle of Wind":
                job_object.ActionSet.append(RiddleOfWind)
            elif aura == "Brotherhood":
                job_object.ActionSet.append(Brotherhood)
            elif aura == "Life Surge":
                job_object.ActionSet.append(LifeSurge)
            elif aura == "Lance Charge":
                job_object.ActionSet.append(LanceCharge)
            elif aura == "Battle Litany":
                job_object.ActionSet.append(BattleLitany)
            elif aura == "Right Eye":
                if not foundLeftEye:
                    foundRightEye = True
                    rightEyeId = job_object.playerID
                else:
                    job_object.ActionSet.append(DragonSight(PlayerActionList[str(leftEyeTargetId)]['job_object']))
            elif aura == "Left Eye":
                if not foundRightEye:
                    foundLeftEye = True
                    leftEyeTargetId = job_object.playerID
                else:
                    PlayerActionList[str(rightEyeId)]['job_object'].ActionSet.append(DragonSight(job_object))
            elif aura == "Kassatsu":
                job_object.ActionSet.append(Kassatsu)
            elif aura == "Meikyo Shisui":
                job_object.ActionSet.append(Meikyo)
            elif aura == "Battle Voice":
                if job_object.JobEnum == JobEnum.Bard:
                    job_object.ActionSet.append(BattleVoice)
            elif aura == "Acceleration":
                job_object.ActionSet.append(Acceleration)
            elif aura == "Embolden":
                if job_object.JobEnum == JobEnum.RedMage:
                    job_object.ActionSet.append(Embolden)
            elif aura == "Searing Light":
                if job_object.JobEnum == JobEnum.Summoner:
                    job_object.ActionSet.append(SearingLight)
            elif aura == "Presence of Mind":
                job_object.ActionSet.append(PresenceOfMind)
            elif 'Drawn' in aura:
                job_object.ActionSet.append(Draw)
            elif aura == "Earthly Dominance":
                job_object.ActionSet.append(EarthlyStar)
            elif aura == "Defiance":
                job_object.ActionSet.append(Defiance)
            elif aura == "Royal Guard":
                job_object.ActionSet.append(RoyalGuard)
            elif aura == "Grit":
                job_object.ActionSet.append(Grit)
            elif aura == "Iron Will":
                job_object.ActionSet.append(IronWill)
            elif aura == "Mudra": # This has nothing to do with Huton but if we see that aura we will give it.
                job_object.ActionSet.append(Huton)
            elif aura == "Reassembled":
                job_object.ActionSet.append(Reassemble)
            elif aura == "Eukrasia":
                job_object.ActionSet.append(Eukrasia)
            elif aura == "Standard Step":
                # Assume 2 steps
                job_object.ActionSet.append(StandardStep)
                job_object.ActionSet.append(Pirouette)
                job_object.ActionSet.append(Jete)
            elif aura == "Technical Step":
                # Assume 4 steps
                job_object.ActionSet.append(TechnicalStep)
                job_object.ActionSet.append(Pirouette)
                job_object.ActionSet.append(Jete)
                job_object.ActionSet.append(Entrechat)
                job_object.ActionSet.append(Emboite)
            elif aura == "Closed Position":
                if not foundDancePartner:
                    foundClosedPosition = True
                    dancerId = job_object.playerID
                else:
                    job_object.ActionSet.insert(0, ClosedPosition(PlayerActionList[str(partnerId)]['job_object']))
            elif aura == "Dance Partner":
                if not foundClosedPosition:
                    foundDancePartner = True
                    partnerId = job_object.playerID
                else:
                    PlayerActionList[str(dancerId)]['job_object'].ActionSet.insert(0, ClosedPosition(job_object))

    foundDancePartner = False
    foundClosedPosition = False
    dancerId = None
    partnerId = None
    #Will now go through every player and give them an ActionList


    for playerID in PlayerActionList:

        player_obj = PlayerActionList[playerID]["job_object"] # Getting job object so we can have access to the ActionEnum of the class

        for action in PlayerActionList[playerID]["actionList"]:

            actionID = (action["abilityGameID"] 
                       if 'abilityGameID' in action.keys() else 
                       id_for_name(action["actionName"], player_obj.ClassAction, player_obj.JobAction)) # Getting id from name

            if int(actionID) == 212 : 
                #WaitAbility. WaitAbility has a special field where the waited time is specified
                actionObject = WaitAbility(float(action["waitTime"]))
            else: 
                if "targetID" in action.keys(): # Action has a target
                    actionObject = lookup_abilityID(actionID,action["targetID"], playerID,PlayerActionList) #Getting action object
                else:
                    actionObject = lookup_abilityID(actionID,0, playerID,PlayerActionList) # Target is Enemy which defaults to id 0

            PlayerActionList[playerID]["actionObject"] += [actionObject]

        #We will now create the event

    Dummy = Enemy()
    Event = Fight(Dummy, False)

    for playerID in PlayerActionList:
        PlayerActionList[playerID]["job_object"].ActionSet += PlayerActionList[playerID]["actionObject"] #Linking player object and action list
        Event.AddPlayer([PlayerActionList[playerID]["job_object"]], giveNewID=False) #Adding job_object to Event

    return Event

def SimulateFightBackend(file_name : str):
    """
    This function takes a file_name and opens the given file name assuming it is located in the saved folder.
    file_name : str -> name of the saved file.
    """

    f = open(file_name) #Opening save

    data = json.load(f) #Loading json file
    PlayerList = data["data"]["PlayerList"]

    if len(PlayerList) > 1 : #If there is more than 1 player we will ask if we wish to simulate all or only 1
        print(
            "This save file has more than one player character : " + "\n" + 
            "1- Simulate only one player character" + "\n" + 
            "2- Simulate all the player characters at the same time" + "\n" + 
            "====================================================="
            )

        user_input = AskInput(2)

        if user_input == "1" : #want to only simulate one player
            print(
                "=====================================================" + "\n" + 
                "List of players in this save file : "
            )
            n_player = len(PlayerList)
            for i in range(n_player):
                print(str(i + 1) + " -> " + PlayerList[i]["JobName"])
            
            print("Select which player character you wish to simulate.")

            user_input = AskInput(n_player)

            PlayerList = [PlayerList[int(user_input) - 1]] #Only taking what we are interested in

    # Asking for how many ZIP simulations
    n = 0
    if (input("Do you want to run ZIPSimulations?(y/n)") == "y"):
        n = int(input("Enter the amount : "))
    

    #We will now go through all actionID and transform into an abilityList

    print("Restoring save file into Event object...")

    Event = RestoreFightObject(data, name=file_name)

    fightInfo = data["data"]["fightInfo"] #fight information
    Event.ShowGraph = fightInfo["ShowGraph"] #Default
    Event.RequirementOn = fightInfo["RequirementOn"]
    Event.IgnoreMana = fightInfo["IgnoreMana"]
    
    Event.SimulateFight(0.01,fightInfo["fightDuration"], vocal=True, n = n) #Simulates the fight


    print(
        "========================================="
        )
    input("Press any key to return to the Main menu : ")

def MergeFightBackEnd(child_fight, parent_fight, parent_name):
    #This will merge the two fights.

    
    child = open(child_fight) #Opening save
    parent = open(parent_fight)
    data_child = json.load(child) #Loading json file
    data_parent = json.load(parent) #Loading json file

    #We will simply put Player list of data_child into player_list of parent

    data_parent["data"]["PlayerList"] += data_child["data"]["PlayerList"]

    save_dir: Path = Path.cwd() / "saved"
    with open(save_dir / parent_name, "w") as write_files:
        json.dump(data_parent,write_files, indent=4) #saving file

def GenerateLayoutBackend(player_list,namefile):
    # This function will generate a JSON file with nothng written inside which can be edited by the user
    # directly. It requires as input a player_list which is a list of PlayerEnum to know which player
    # the user wants in the fight

    data = GenerateLayoutDict(player_list)
    
    save_dir: Path = Path.cwd() / 'saved'
    with open(save_dir / f'{namefile}.json', "w") as write_files:
        json.dump(data,write_files, indent=4) # saving file

    
def GenerateLayoutDict(player_list):
    """
    This function generates a dictionnary that the simulator can use to simulate the fight
    """

    data = {
        "data":
        { 
            "fightInfo"  : {
                "fightDuration" : 500,
                "time_unit" : 0.01,
                "ShowGraph" : True,
                "RequirementOn" : True,
                "IgnoreMana" : False
            },
            "PlayerList" : []
            }
    }

    stat_dict = {
            "MainStat" : 390,
            "WD" : 0,
            "Det" : 390,
            "Ten" : 400,
            "SS" : 400,
            "SkS" : 400,
            "Crit" : 400,
            "DH" : 400
        }

    # Will not fill the PlayerList
    id = 1

    for player in player_list:
        # player is a JobEnum

        player_dict = {
            "JobName" : JobEnum.name_for_id(player),
            "playerID" : id,
            "stat" : stat_dict,
            "etro_gearset_url":  "Put a URL here if you want. The code will only overwirte the stats if it detects an etro url.",
            "Auras" : [],
            "actionList" : [
                {"actionName": "putNameHere"}
            ]
        }

        data["data"]["PlayerList"].append(copy.deepcopy(player_dict))

        id+=1

    return data

# Condition functions to go with conditionalAction

def timeCheck(timeStamp : float):
    """
    This function returns a condition function to be used with conditionalAction.
    The returned function will return True if the timeStamp of the fight is less than the given timeStamp.
    False otherwise.
    timeStamp : float -> timeStamp to compare against. Function returns true if the timeStamp of the fight
                         is less than this timeStamp and False otherwise.
    """

    def __condition(Player):
        return Player.CurrentFight.TimeStamp < timeStamp
    return __condition


from ffxivcalc.Jobs.Base_Spell import DOTSpell, WaitAbility, buff, empty, buffPercentHistory
from ffxivcalc.Jobs.Melee.Melee_Spell import NinjaSpell
import copy

from ffxivcalc.Jobs.Player import Pet

Lock = 0


#Requirement

def ShadeShiftRequirement(Player, Spell):
    return Player.ShadeShiftCD <= 0, Player.ShadeShiftCD

def PhantomKamaitachiRequirement(Player, Spell):
    return Player.PhantomKamaitachiReady, -1

def BunshinRequirement(Player, Spell):
    return Player.BunshinCD <=0 and Player.NinkiGauge >= 50, Player.BunshinCD

def TenRequirement(Player, Spell):
    if Player.Ten :
        Player.Ten = False
        return True, -1
    return False, -1

def ChiRequirement(Player, Spell):
    return Player.Chi, -1

def JinRequirement(Player, Spell):
    return Player.Jin, -1

def TenChiJinRequirement(Player, Spell):
    return Player.TenChiJinCD <= 0 and not Player.Kassatsu, Player.TenChiJinCD

def KassatsuOnRequirement(Player, Spell):
    return Player.Kassatsu, -1

def KassatsuRequirement(Player, Spell):
    return Player.KassatsuCD <= 0, Player.KassatsuCD

def NinjutsuRequirement(Player, Spell):
    return Player.NinjutsuStack > 0 or Player.Kassatsu or len(Player.CurrentRitual) > 0, Player.NinjutsuCD

def FleetingRaijuRequirement(Player, Spell):
    return Player.RaijuStack > 0, -1

def MeisuiRequirement(Player, Spell):
    return Player.Suiton and Player.MeisuiCD <= 0, -1

def BhavacakraRequirement(Player, Spell):
    return Player.NinkiGauge >= 50, -1

def TrickAttackRequirement(Player, Spell):
    return Player.Suiton and Player.TrickAttackCD <= 0, -1

def MugRequirement(Player, Spell):
    return Player.MugCD <= 0, Player.MugCD

def DreamWithinADreamRequirement(Player, Spell):
    return Player.DreamWithinADreamCD <= 0, Player.DreamWithinADreamCD

def TenChiJinOnRequirement(Player, Spell): #TenChiJin Ninjutsu
    return Player.TenChiJinTimer > 0, -1

#Ninjutsu Requirement

def FumaShurikenRequirement(Player, Spell):
    return Player.CurrentRitual == [0] or Player.CurrentRitual == [1] or Player.CurrentRitual == [2], -1

def RaitonRequirement(Player, Spell):
    return Player.CurrentRitual == [0,1] or Player.CurrentRitual == [2,1], -1

def KatonRequirement(Player, Spell):
    return Player.CurrentRitual == [1,0] or Player.CurrentRitual == [2,0], -1

def HyotonRequirement(Player, Spell):
    return Player.CurrentRitual == [0,2] or Player.CurrentRitual == [1,2], -1

def HutonRequirement(Player, Spell):
    return Player.CurrentRitual == [2,1,0] or Player.CurrentRitual == [1,2,0], -1

def DotonRequirement(Player, Spell):
    return Player.CurrentRitual == [0,2,1] or Player.CurrentRitual == [2,0,1], -1

def SuitonRequirement(Player, Spell):
    return Player.CurrentRitual == [0,1,2] or Player.CurrentRitual == [1,0,2], -1


#Apply

def ApplyDeathBlossom(Player, Enemy):
    if not (DeathBlossomCombo in Player.EffectList) : Player.EffectList.append(DeathBlossomCombo)
    Player.AddNinki(5)

def ApplyShadeShift(Player, Enemy):
    Player.ShadeShiftCD = 120

def ApplyDoton(Player, Enemy):
    Player.ResetRitual()
    Player.DotonTimer = 18
    if Player.DotonDOT == None:
        Player.DotonDOT = copy.deepcopy(DotonDOT)
        Player.DOTList.append(Player.DotonDOT)
        Player.EffectCDList.append(DotonCheck)
    else : Player.DotonDOT.resetBuffSnapshot()

def ApplyTen(Player, Enemy):
    ApplyNinjutsu(Player, Enemy)
    Player.CurrentRitual.append(0) #0 is a Ten

def ApplyChi(Player, Enemy):
    ApplyNinjutsu(Player, Enemy)
    Player.CurrentRitual.append(1) #1 is a Chi

def ApplyJin(Player, Enemy):
    ApplyNinjutsu(Player, Enemy)
    Player.CurrentRitual.append(2) #2 is a Jin

def ApplyHide(Player, Enemy):
    Player.NinjutsuStack = 2
    if Player.NinjutsuCD > 0:
        Player.EffectCDList.remove(NinjutsuStackCheck)
        Player.NinjutsuCD = 0

def ApplyPhantomKamaitachi(Player, Enemy):

    if Player.Pet == None: #If no Shadow, we will create a new one
        Player.Pet = Pet(Player) # Creating an object
        Player.Pet.ActionSet.append(WaitAbility(0.01)) #Adding an action so program does not crash

    Player.PhantomKamaitachiReady = False
    #Will give Action to the Shadow
    Shadow = Player.Pet
    Shadow.TrueLock = False #Delocking the shadow
    Shadow.ActionSet.insert(Shadow.NextSpell + 1, PhantomKamaitachiShadow) #Adding the spell
    Player.AddHuton(10)
    Player.AddNinki(10)

def ApplyBunshin(Player, Enemy):
    Player.AddNinki(-50)
    Player.BunshinCD = 90
    Player.BunshinStack = 5

    if Player.Pet == None: #If no Shadow, we will create a new one
        Player.Pet = Pet(Player) # Creating an object
        Player.Pet.ActionSet.append(WaitAbility(0.01)) #Adding an action so program does not crash
    else:
        Player.Pet.ResetStat() # Reseting stat on the summon

    Player.EffectList.append(BunshinEffect)
    Player.PhantomKamaitachiReady = True
    Player.PhantomKamaitachiReadyTimer = 45
    Player.EffectCDList.append(PhantomKamaitachiCheck)

def ApplyTenChiJin(Player, Enemy):
    Player.TenChiJinCD = 120
    Player.TenChiJinTimer = 6
    Player.EffectCDList.append(TenChiJinCheck)
    Player.EffectList.append(TenChiJinEffect)

def ApplyKassatsu(Player, Enemy):
    Player.KassatsuTimer = 15
    Player.KassatsuCD = 60
    Player.Kassatsu = True
    Player.EffectList.append(KassatsuEffect)
    Player.EffectCDList.append(KassatsuCheck)

def ApplyHyoshoRanryu(Player, Enemy):
    Player.ResetRitual()

def ApplySuiton(Player, Enemy):
    Player.Suiton = True
    if not SuitonCheck in Player.EffectCDList : Player.EffectCDList.append(SuitonCheck)
    Player.SuitonTimer = 20
    Player.ResetRitual() #Even if in TenChiJin, does not matter

def ApplyHuton(Player, Enemy):
    Player.HutonTimer = 60
    if not (HutonEffect in Player.EffectList):
        Player.EffectList.append(HutonEffect)
    if not (HutonCheck in Player.EffectCDList):
        Player.Haste += 15
        Player.hasteChangeValue = 15
        Player.hasteHasChanged = True
        Player.EffectCDList.append(HutonCheck)
    Player.ResetRitual() #Even if in TenChiJin, does not matter

def ApplyRaiton(Player, Enemy):
    if Player.RaijuStack == 0: Player.EffectList.append(RaitonEffect) #will loose all if weaponskill is done
    Player.RaijuStack = min(3, Player.RaijuStack + 1)
    Player.ResetRitual() #Even if in TenChiJin, does not matter

def ApplyNinjutsu(Player, Enemy):
    if len(Player.CurrentRitual) == 0 and not Player.Kassatsu: #If we are not already in a ritual
        if Player.NinjutsuStack == 2:
            Player.EffectCDList.append(NinjutsuStackCheck)
            Player.NinjutsuCD = 20
        Player.NinjutsuStack -= 1

def ApplyThrowingDagger(Player, Enemy):
    Player.AddNinki(5)

def ApplyFleetingRaiju(Player, Enemy):
    Player.RaijuReady = False
    Player.RaijuStack -= 1
    Player.AddNinki(5)
    

def ApplyMeisui(Player, Enemy):
    Player.Suiton = False
    Player.AddNinki(50)
    Player.MeisuiCD = 120
    Player.MeisuiTimer = 30
    Player.EffectList.append(MeisuiEffect)
    Player.EffectCDList.append(MeisuiCheck)

def ApplyBhavacakra(Player, Enemy):
    Player.AddNinki(-50)

def ApplyTrickAttack(Player, Enemy):
    Player.TrickAttackCD = 60
    Player.TrickAttackTimer = 15
    Player.EffectCDList.append(TrickAttackCheck)
    Player.buffList.append(TrickAttackBuff)
    Player.Suiton = False

                             # Only relevant to PreBakedAction and only does that code if true. Also checking if the PreBakedAction player is this ninja
    #if Player.CurrentFight.SavePreBakedAction and Player == Player.CurrentFight.PlayerList[Player.CurrentFight.PlayerIDSavePreBakedAction]:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 15, TrickAttackBuff.MultDPS)
    #    history.isTrickAttack = True
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplyMug(Player, Enemy):
    Player.MugCD = 120
    Player.MugTimer = 20
    Player.AddNinki(40)
    Player.EffectCDList.append(MugCheck)
    Enemy.buffList.append(MugBuff)

    # Only relevant to PreBakedAction and only does that code if true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20, MugBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplyHuraijin(Player, Enemy):
    Player.HutonTimer = 60
    if not (HutonEffect in Player.EffectList):
        Player.EffectList.append(HutonEffect)
    if not (HutonCheck in Player.EffectCDList):
        Player.Haste += 15
        Player.hasteChangeValue = 15
        Player.hasteHasChanged = True
        Player.EffectCDList.append(HutonCheck)

def ApplyDreamWithinADream(Player, Enemy):
    Player.DreamWithinADreamCD = 60

def ApplySpinningEdge(Player, Enemy):
    ##input(Player.EffectList)
    Player.AddNinki(5)
    
    if not (SpinningEdgeCombo in Player.EffectList): Player.EffectList.append(SpinningEdgeCombo)


#Effect

def DeathBlossomCombo(Player, Spell):
    if Spell.id == HakkeMujinsatsu.id:
        Spell.Potency += 30
        Player.AddHuton(10)
        Player.AddNinki(5)
        Player.EffectToRemove.append(DeathBlossomCombo)

def TenChiJinEffect(Player, Spell):
    if Spell.id == Ten2.id or Spell.id == Chi2.id or Spell.id == Jin2.id: #If not one of these spells, we simply stop

        if Spell.id == Ten2.id:
            Player.TenChiJinRitual += [0] #Ten is 0
        elif Spell.id == Chi2.id:
            Player.TenChiJinRitual += [1] #Chi is 1
        elif Spell.id == Jin2.id:
            Player.TenChiJinRitual += [2] #Jin is 2

        #We will now check how we change the action

        if len(Player.TenChiJinRitual) == 1: #We will execute FumaShuriken
            Spell.Potency += FumaShuriken.Potency #Giving Potency as a bonus in case some other effect gave flat potency (shouldn't be the case for Ninja)
            #No notable effect to give
        elif len(Player.TenChiJinRitual) == 2:

            if Player.TenChiJinRitual == [0,1] or Player.TenChiJinRitual == [2,1]: # Raiton : Ten -> Chi or Jin -> Chi
                Spell.Potency += Raiton.Potency
                Spell.Effect = Raiton.Effect
            elif Player.TenChiJinRitual == [0, 2] or Player.TenChiJinRitual == [1,2]: # hyoton : Ten -> Jin or Chi -> Jin
                Spell.Potency += Hyoton.Potency
                #No notable effect to give
            elif Player.TenChiJinRitual == [1,0] or Player.TenChiJinRitual == [2,0]: # Katon : Chi -> Ten or Jin -> Ten
                Spell.Potency += Katon.Potency
                #No notable effect to give
            #Else, we change nothing and its going to bunny lol
        
        elif len(Player.TenChiJinRitual) == 3:

            if Player.TenChiJinRitual == [0,1,2] or Player.TenChiJinRitual == [1,0,2]: # Suiton : Ten -> Chi -> Jin or Chi -> Ten -> Jin
                Spell.Potency += Suiton.Potency
                Spell.Effect = Suiton.Effect
                Spell.RecastTime += 0.5
            elif Player.TenChiJinRitual == [2,1,0] or Player.TenChiJinRitual == [1,2,0]: # Huton : Jin -> Chi -> Ten or Chi -> Jin -> Ten
                Spell.Potency = 0
                Spell.Effect = Huton.Effect
            elif Player.TenChiJinRitual == [0,2,1] or Player.TenChiJinRitual == [2,0,1]: # Doton : Ten -> Jin -> Chi or Jin -> Ten -> Chi
                Spell.Potency = 0
                Spell.Effect = Doton.Effect

            #Since last possible Ninjutsu, we will set TenChiJin Timer to 0.01 so it terminates
            Player.TenChiJinTimer = 0.01
            Player.TenChiJinRitual = [] #Reseting TenChiJin Ritual
            #input("Finishing")
            #Effect will be removed in Check

            


    elif not (isinstance(Spell, DOTSpell)): #If not a DOTSpell
        Player.TenChiJinTimer = 0


def HutonEffect(Player, Spell):
    pass
        # Why did I do that? Left for future me to figure out
    #if isinstance(Spell, NinjaSpell) and Spell.Weaponskill and Spell.GCD : Spell.RecastTime *= 0.85

def BunshinEffect(Player, Spell):
    if isinstance(Spell, NinjaSpell) and Spell.Weaponskill and Spell.id != PhantomKamaitachi.id: #We don't want Kamataichi to have bunshin effect
        #Will give Action to the Shadow
        Shadow = Player.Pet
        Shadow.TrueLock = False #Delocking the shadow
        Shadow.ActionSet.insert(Shadow.NextSpell + 1, NinjaSpell(35, False, 0, 1, 160, empty, [], True, False)) #Adding the spell
        Player.BunshinStack -= 1
        Player.AddNinki(5)
        if Player.BunshinStack == 0:
            Player.EffectToRemove.append(BunshinEffect)

def KassatsuEffect(Player, Spell):
    if isinstance(Spell, NinjaSpell) and Spell.Ninjutsu:
        Spell.DPSBonus = 1.3
        Player.KassatsuTimer = 0

def RaitonEffect(Player, Spell):
    if Spell.GCD and (Spell.Weaponskill or Player.RaijuStack == 0) and (Spell.id != FleetingRaiju.id and Spell.id != PhantomKamaitachi.id):
        Player.RaijuStack = 0
        Player.EffectToRemove.append(RaitonEffect)

def MeisuiEffect(Player, Spell):
    if Spell.id == Bhavacakra.id:
        Spell.Potency += 150

def SpinningEdgeCombo(Player, Spell):
    if Spell.id == GustSlash.id:
        Spell.Potency += 160
        Player.AddNinki(5)
        Player.EffectToRemove.append(SpinningEdgeCombo)
        if not (GustSlashCombo in Player.EffectList) : Player.EffectList.append(GustSlashCombo)

def GustSlashCombo(Player, Spell):
    if Spell.id == AeolianEdge.id:
        Spell.Potency += 240
        Player.AddNinki(15)
        Player.EffectToRemove.append(GustSlashCombo)
    elif Spell.id == ArmorCrush.id:
        Spell.Potency += 220
        Player.AddNinki(15)
        Player.AddHuton(30)
        Player.EffectToRemove.append(GustSlashCombo)

#Check

def DotonCheck(Player, Enemy):
    if Player.DotonTimer <= 0:
        Player.DOTList.remove(Player.DotonDOT)
        Player.DotonDOT = None
        Player.EffectToRemove.append(DotonCheck)

def HutonCheck(Player, Enemy):
    if Player.HutonTimer <= 0:
        Player.EffectList.remove(HutonEffect)
        Player.EffectToRemove.append(HutonCheck)
        Player.Haste -= 15
        Player.hasteChangeValue = -15
        Player.hasteHasChanged = True

def PhantomKamaitachiCheck(Player, Enemy):
    if not Player.PhantomKamaitachiReady or Player.PhantomKamaitachiReadyTimer <= 0:
        Player.PhantomKamaitachiReady = False
        Player.PhantomKamaitachiReadyTimer = 0
        Player.EffectToRemove.append(PhantomKamaitachiCheck)

def SuitonCheck(Player, Enemy):
    if not Player.Suiton or Player.SuitonTimer <= 0:
        Player.EffectToRemove.append(SuitonCheck)
        Player.SuitonTimer = 0
        Player.Suiton = False

def KassatsuCheck(Player, Enemy):
    if Player.KassatsuTimer <= 0:
        Player.EffectList.remove(KassatsuEffect)
        Player.EffectToRemove.append(KassatsuCheck)
        Player.Kassatsu = False

def NinjutsuStackCheck(Player, Enemy):
    if Player.NinjutsuCD <= 0:
        if Player.NinjutsuStack == 1:
            Player.EffectToRemove.append(NinjutsuStackCheck)
        else:
            Player.NinjutsuCD = 20
        Player.NinjutsuStack +=1

def MeisuiCheck(Player, Enemy):
    if Player.MeisuiTimer <= 0:
        Player.EffectList.remove(MeisuiEffect)
        Player.EffectToRemove.append(MeisuiCheck)

def TrickAttackCheck(Player, Enemy):
    if Player.TrickAttackTimer <= 0:
        Player.buffList.remove(TrickAttackBuff)
        Player.EffectToRemove.append(TrickAttackCheck)

def MugCheck(Player, Enemy):
    if Player.MugTimer <= 0:
        Enemy.buffList.remove(MugBuff)
        Player.EffectToRemove.append(MugCheck)

def TenChiJinCheck(Player, Enemy):
    if Player.TenChiJinTimer <= 0:
        Player.EffectList.remove(TenChiJinEffect)
        Player.EffectToRemove.append(TenChiJinCheck)



#GCD
SpinningEdge = NinjaSpell(2240, True, Lock, 2.5, 220, ApplySpinningEdge, [], True, False, type = 2)
GustSlash = NinjaSpell(2242, True, Lock, 2.5, 160, empty, [], True, False, type = 2)
AeolianEdge = NinjaSpell(2255, True, Lock, 2.5, 200, empty, [], True, False, type = 2)
ArmorCrush = NinjaSpell(3563, True, Lock, 2.5, 200, empty, [], True, False, type = 2)
Huraijin = NinjaSpell(25876, True, Lock, 2.5, 200, ApplyHuraijin, [], True, False, type = 2)
FleetingRaiju = NinjaSpell(25777, True, Lock, 2.5, 560, ApplyFleetingRaiju, [FleetingRaijuRequirement], True, False, type = 2)
ThrowingDagger = NinjaSpell(2247, True, Lock, 2.5, 120, ApplyThrowingDagger, [], True, False, type = 2)
PhantomKamaitachi = NinjaSpell(25774, True, Lock, 2.5, 0, ApplyPhantomKamaitachi, [PhantomKamaitachiRequirement], True, False, type = 2)
PhantomKamaitachiShadow = NinjaSpell(8, False, 0, 0, 600, empty, [], False, False) #Action done by the shadow
DeathBlossom = NinjaSpell(2254, True, Lock, 2.5, 100, ApplyDeathBlossom, [],True, False, type = 2)
HakkeMujinsatsu = NinjaSpell(16488, True, Lock, 2.5, 100, empty, [], True, False, type = 2)

#Ninjutsu
FumaShuriken = NinjaSpell(2265, True, Lock,1.5, 450, ApplyHyoshoRanryu, [FumaShurikenRequirement], False, True) #Same effect as HyoshoRanruy, since only reset Player.CurrentRitual list
Raiton = NinjaSpell(2267, True, Lock,1.5, 650, ApplyRaiton, [RaitonRequirement], False, True )
Huton = NinjaSpell(2269, True, Lock,1.5, 0, ApplyHuton, [HutonRequirement], False, True)
Suiton = NinjaSpell(2271, True, Lock,1.5, 500, ApplySuiton, [SuitonRequirement], False, True)
Hyoton = NinjaSpell(2268, True, Lock, 1.5, 350, ApplyHyoshoRanryu, [HyotonRequirement], False, True)
HyoshoRanryu = NinjaSpell(16492, True, Lock,1.5, 1300, ApplyHyoshoRanryu, [HyotonRequirement, KassatsuOnRequirement], False, True)
Katon = NinjaSpell(2266, True, Lock, 1.5, 350, ApplyHyoshoRanryu, [KatonRequirement], False, True)
GokaMekkyaku = NinjaSpell(16491, True, Lock, 1.5, 600, ApplyHyoshoRanryu, [KatonRequirement, KassatsuOnRequirement], False, True)
Doton = NinjaSpell(2270, True, Lock, 1.5, 0, ApplyDoton, [DotonRequirement], False, True)
DotonDOT = DOTSpell(-33, 80, True, isGround=True)

#Ritual
Ten = NinjaSpell(2259, True, 0, 0.5, 0, ApplyTen, [NinjutsuRequirement], False, False)
Chi = NinjaSpell(2261, True, 0, 0.5, 0, ApplyChi, [NinjutsuRequirement], False, False)
Jin = NinjaSpell(2263, True, 0, 0.5, 0, ApplyJin, [NinjutsuRequirement], False, False)

#TenChiJin
TenChiJin = NinjaSpell(7403, False, Lock, 0, 0, ApplyTenChiJin, [TenChiJinRequirement], False, False)
#TenChiJin will trigger the TenChiJinEffect. The actions we can take will be different from the Ninjutsu
Ten2 = NinjaSpell(18873, True, Lock, 1, 0, empty, [TenChiJinOnRequirement], False, True)
Chi2 = NinjaSpell(18877, True, Lock, 1, 0, empty, [TenChiJinOnRequirement], False, True)
Jin2 = NinjaSpell(18881, True, Lock, 1, 0, empty, [TenChiJinOnRequirement], False, True) 
#These Ninjutsu are only to be used in TenChiJin
#Their potency/apply will be changed according to what spell should be casted

#oGCD
DreamWithinADream = NinjaSpell(3566, False, Lock, 0, 3*150, ApplyDreamWithinADream, [DreamWithinADreamRequirement], False, False)
Mug = NinjaSpell(2248, False, Lock, 0, 150, ApplyMug, [MugRequirement], False, False)
TrickAttack = NinjaSpell(2258, False, Lock, 0, 400, ApplyTrickAttack, [TrickAttackRequirement], False, False)
Bhavacakra = NinjaSpell(7402, False, Lock, 0, 350, ApplyBhavacakra, [BhavacakraRequirement], False, False)
Meisui = NinjaSpell(16489, False, Lock, 0, 0, ApplyMeisui, [MeisuiRequirement], False, False)
Kassatsu = NinjaSpell(2264, False, Lock, 0, 0, ApplyKassatsu,[KassatsuRequirement], False, False)
Bunshin = NinjaSpell(16493, False, Lock, 0, 0, ApplyBunshin, [BunshinRequirement], False, False)
Hide = NinjaSpell(2245, True, 0, 0, 0, ApplyHide, [], False, False)
ShadeShift = NinjaSpell(2241, False, Lock, 0, 0, ApplyShadeShift, [ShadeShiftRequirement], False, False)
HellfrogMedium = NinjaSpell(7401, False, Lock, 0, 160, ApplyBhavacakra, [BhavacakraRequirement], False, False)
#buff
MugBuff = buff(1.05,name="Mug", isDebuff = True)
TrickAttackBuff = buff(1.1,name="Trick Attack", isDebuff = True)


NinjaAbility = {
2240 : SpinningEdge,
2242 : GustSlash,
2255 : AeolianEdge,
3563 : ArmorCrush,
2247 : ThrowingDagger,
2241 : ShadeShift,
2254 : DeathBlossom,
16488 : HakkeMujinsatsu,
2245 : Hide,
2248 : Mug,
2258 : TrickAttack,
2259 : Ten,
2261 : Chi,
2263 : Jin,
2265 : FumaShuriken,
2266 : Katon,
2267 : Raiton,
2268 : Hyoton,
2269 : Huton,
2270 : Doton,
2271 : Suiton,
16491 : GokaMekkyaku,
16492: HyoshoRanryu,
25774 : PhantomKamaitachi,
#25776 : Hollow Nozuchi,
3566 : DreamWithinADream,
25876 : Huraijin,
2264 : Kassatsu,
2262 : WaitAbility(0.5), #Shukuchi,
7401 : HellfrogMedium,
7402 : Bhavacakra,
7403 : TenChiJin,
16489 : Meisui,
16493 : Bunshin,
25777 : FleetingRaiju,
25778 : FleetingRaiju,
18805 : Ten,
18806 : Chi,
18807 : Jin,
18873 : Ten2,
18877 : Chi2,
18881 : Jin2
}
